"""Implementation of generic SHV RPC broker."""

import argparse
import asyncio
import logging
import pathlib

from .broker import RpcBroker
from .config import RpcBrokerConfig

logger = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    """Parse passed arguments and return result."""
    parser = argparse.ArgumentParser("pyshvbroker", description="Silicon Heaven broker")
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {RpcBroker.Client.APP_VERSION}",
    )
    parser.add_argument(
        "-v",
        action="count",
        default=0,
        help="Increase verbosity level of logging",
    )
    parser.add_argument(
        "-q",
        action="count",
        default=0,
        help="Decrease verbosity level of logging",
    )
    levels = [level for level in logging.getLevelNamesMapping() if level != "NOTSET"]
    parser.add_argument(
        "--log-level",
        action="store",
        choices=levels,
        help="Set logging level exactly to specific level (ignores -v and -q)."
        + f"The default level is WARNING. Supported levels: {', '.join(levels)}",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        default="/etc/pyshvbroker.toml",
        type=pathlib.Path,
        help="Configuration file",
    )
    return parser.parse_args()


async def _broker_main(config: RpcBrokerConfig) -> None:
    broker = RpcBroker(config)
    try:
        await broker.serve_forever()
    finally:
        await broker.terminate()


def main() -> None:
    """Application's entrypoint."""
    args = parse_args()

    logging.basicConfig(
        level=logging.getLevelNamesMapping()[args.log_level]
        if args.log_level
        else logging.WARN + ((args.q - args.v) * 10),
        format="[%(asctime)s] [%(levelname)s] - %(message)s",
    )

    brokerconf = RpcBrokerConfig.load(args.config)
    try:
        asyncio.run(_broker_main(brokerconf))
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



"""Check that we can serialize and deserialize Chainpack."""

import datetime
import decimal

import pytest

from shv import ChainPackReader, ChainPackWriter, SHVMeta, SHVUInt, shvmeta

# You can get chainpack using this shell command:
#   echo 'null' | cp2cp --ip --oc \
#     | python -c "import sys; print(repr(sys.stdin.buffer.read()))"
DATA: list = [
    (b"\x80", None),
    (b"\xfe", True),
    (b"\xfd", False),
    (b"@", 0),
    (b"\x82A", -1),
    (b"\x82B", -2),
    (b"G", 7),
    (b"j", 42),
    (b"\x01", SHVUInt(1)),
    (b"\x81\xf0\x7f\xff\xff\xff", SHVUInt(2**31 - 1)),
    (b"\x81\xf0\xff\xff\xff\xff", SHVUInt(2**32 - 1)),
    (b"\x81\xf4\x80\x00\x00\x00\x00\x00\x00\x00", SHVUInt(1 << 63)),
    (b"\x82\xf0\x7f\xff\xff\xff", 2**31 - 1),
    (b"\x82\xf1\x00\xff\xff\xff\xff", 2**32 - 1),
    (b"\x82\xf3\x1f\xff\xff\xff\xff\xff\xff", 2**53 - 1),
    (b"\x82\xf3\x9f\xff\xff\xff\xff\xff\xff", 1 - 2**53),
    (b"\x83\x00\x00\x00\x00\x00\xe0k@", 223.0),
    (b"\x8c\x17A", decimal.Decimal((0, (2, 3), -1))),
    (b"\x86\x00", ""),
    (b"\x86\x03foo", "foo"),
    (b"\x88\xff", []),
    (b"\x88A\xff", [1]),
    (b"\x88ABC\xff", [1, 2, 3]),
    (b"\x88\x88\xff\xff", [[]]),
    (b"\x89\x86\x03foo\x86\x03bar\xff", {"foo": "bar"}),
    (b"\x8aAB\xff", {1: 2}),
    (
        b"\x88\x01\x89\x86\x01aA\xff\x8c\x17A\xff",
        [SHVUInt(1), {"a": 1}, decimal.Decimal("2.3")],
    ),
    (b"\x8bAB\xffC", SHVMeta.new(3, {1: 2})),
    (b"\x88A\x8bGH\xffI\xff", [1, SHVMeta.new(9, {7: 8})]),
    (
        b"\x8bH\x03\xff\x8aB\x88\x88\x86\x07.broker\x8bAB\xff\xfe\xff\xff\xff",
        SHVMeta.new({2: [[".broker", SHVMeta.new(True, {1: 2})]]}, {8: SHVUInt(3)}),
    ),
    (
        b"\x8bAB\x86\x03foo\x8bEF\xff\x86\x03bar\xff\x88\x01\x89\x86\x01aA\xff\x8c\x17A\xff",
        SHVMeta.new(
            [SHVUInt(1), {"a": 1}, decimal.Decimal("2.3")],
            {1: 2, "foo": SHVMeta.new("bar", {5: 6})},
        ),
    ),
    (
        b"\x8bAB\xff\x88C\x8bDE\xffF\xff",
        SHVMeta.new([3, SHVMeta.new(6, {4: 5})], {1: 2}),
    ),
    (
        b"\x8bD\x86\x05svete\xff\x8aB\x8bD\x86\x05svete\xff\x88@A\xff\xff",
        SHVMeta.new({2: SHVMeta.new([0, 1], {4: "svete"})}, {4: "svete"}),
    ),
    (b"\x85\x06ab\xcd\t\r\n", b"ab\xcd\t\r\n"),
    (
        b"\x8d\x04",
        datetime.datetime(2018, 2, 2, 0, 0, 0, 1000, tzinfo=datetime.UTC),
    ),
    (
        b"\x8d\x82\x11",
        datetime.datetime(
            2018,
            2,
            2,
            1,
            0,
            0,
            1000,
            tzinfo=datetime.timezone(datetime.timedelta(hours=1)),
        ),
    ),
]


@pytest.mark.parametrize(
    "chainpack,data",
    [
        *DATA,
        (b"\x8efoo\x00", "foo"),
    ],
)
def test_reader(chainpack, data):
    obj = ChainPackReader.unpack(chainpack)
    assert obj == data
    assert shvmeta(obj) == shvmeta(data)


def test_reader_uint():
    assert isinstance(ChainPackReader.unpack(b"\x01"), SHVUInt)


@pytest.mark.parametrize(
    "chainpack,data",
    [
        *DATA,
        (b"A", SHVMeta.new(1)),
    ],
)
def test_writer(chainpack, data):
    res = ChainPackWriter.pack(data)
    assert res == chainpack


def test_writer_cstring():
    obj = ChainPackWriter()
    obj.write_cstring("foo")
    assert obj.stream.getvalue() == b"\x8efoo\x00"



"""Tools to parse common SHV parameters."""

from __future__ import annotations

import collections.abc
import typing

from .rpcerrors import RpcInvalidParamError
from .value import SHVType, is_shvlist

SHVT = typing.TypeVar("SHVT", bound=SHVType)


def shvt(value: SHVType, tp: type[SHVT]) -> SHVT:
    """Check type of value.

    This is just simple type check of parameter for simple types.

    :param value: The value received as parameter for SHV RPC.
    :param tp: Python type the value should be of.
    :return: The value.
    :raises RpcInvalidParamError: If doesn't match what was provided in value.
    """
    if not isinstance(value, tp):
        raise RpcInvalidParamError(f"Invalid type: {type(value)}")
    return value


class NoDefaultType:
    """Type for :data:`NO_DEFAULT` singleton."""

    def __new__(cls) -> NoDefaultType:
        """Implement as singleton."""
        if not hasattr(cls, "instance"):
            cls.instance = super(NoDefaultType, cls).__new__(cls)  # noqa UP008
        return cls.instance


NO_DEFAULT = NoDefaultType()
"""Singleton used to identify that there is no default."""


class SHVGetKey(typing.NamedTuple):
    """Key used in :func:`shvget` to allow both IMap as well as Map."""

    skey: str | None = None
    ikey: int | None = None


def shvget(
    value: SHVType,
    key: str | int | SHVGetKey | collections.abc.Sequence[str | int | SHVGetKey],
    default: SHVType | NoDefaultType = NO_DEFAULT,
) -> SHVType:
    """Get value from (i)map or (i)map of (i)maps or default.

    It is very common to query SHVType for keys and support both IMap as well as
    Map at the same time. This can be any number of nested Maps or IMaps. This
    simplifies access for such types.

    :param value: Some Map or IMap to be accessed
    :param key: Key or list of keys that should be recursively applied to the
      value.
    :param default: Default value used if value is not present.
      :class:`RpcInvalidParamError` is raised instead if :data:`NO_DEFAULT` is
      specifiedl.
    :return: extracted value or default.
    :raise RpcInvalidParamError: if the value is not present and no default was
      provided.
    """
    for k in [key] if isinstance(key, str | int | SHVGetKey) else key:
        if not isinstance(value, collections.abc.Mapping):
            break
        vmap = typing.cast(collections.abc.Mapping[str | int, SHVType], value)
        if isinstance(k, SHVGetKey) and k.ikey is not None and k.ikey in vmap:
            value = vmap[k.ikey]
        elif isinstance(k, SHVGetKey) and k.skey is not None and k.skey in vmap:
            value = vmap[k.skey]
        elif isinstance(k, str | int) and k in vmap:
            value = vmap[k]
        else:
            if isinstance(default, NoDefaultType):
                raise RpcInvalidParamError(f"Missing key: {k}")
            value = default
            break
    if value is None and not isinstance(default, NoDefaultType):
        return default  # None is alias for not present for us
    return value


def shvgett(
    value: SHVType,
    key: str | int | SHVGetKey | collections.abc.Sequence[str | int | SHVGetKey],
    tp: type[SHVT],
    default: SHVT | NoDefaultType = NO_DEFAULT,
) -> SHVT:
    """Variant of :func:`shvget` that also checks for type.

    It is very common to query SHVType for keys and expecting a specific type or
    using default value if it is present or is of invalid type. It is a fails
    safe approach.

    :param value: Some Map or IMap to be accessed
    :param key: Key or list of keys that should be recursively applied to the
      value.
    :param tp: Type of the expected value.
    :param default: Default value used if value is not present.
      :class:`RpcInvalidParamError` is raised instead if :data:`NO_DEFAULT` is
      specifiedl.
    :return: extracted value or default.
    :raise RpcInvalidParamError: if the value is not present and no default
      provided.
    """
    return shvt(shvget(value, key, default), tp)


def shvarg(
    value: SHVType,
    index: int,
    default: SHVType | NoDefaultType = NO_DEFAULT,
) -> SHVType:
    """Get value from list or default.

    Some methods expect sequence of values (tuples) as their parameters. This
    helps with parsing such value.

    There is special exception that if the argument is not list then it is
    considered to be the first argument in the list.

    :param value: Some List to be sanitized.
    :param index: The index in list we want to access.
    :param default: Default value used if value is not present.
      :class:`RpcInvalidParamError` is raised instead if :data:`NO_DEFAULT` is
      specifiedl.
    :return: extracted value or default.
    :raise RpcInvalidParamError: if the value is not present and no default
      provided.
    """
    if is_shvlist(value):
        if len(value) > index and (res := value[index]) is not None:
            return res
    elif index == 0 and value is not None:
        return value  # Covers that first argument is sent outside list
    if isinstance(default, NoDefaultType):
        raise RpcInvalidParamError(f"Field {index} not provided")
    return default


def shvargt(
    value: SHVType,
    index: int,
    tp: type[SHVT],
    default: SHVType | NoDefaultType = NO_DEFAULT,
) -> SHVT:
    """Variant of :func:`shvarg` that also checks for type.

    :param value: Some List to be sanitized.
    :param index: The index in list we want to access.
    :param tp: Type of the expected value.
    :param default: Default value used if value is not present.
      :class:`RpcInvalidParamError` is raised instead if :data:`NO_DEFAULT` is
      specifiedl.
    :return: extracted value or default.
    :raise RpcInvalidParamError: if the value is not present and no default
      provided.
    """
    return shvt(shvarg(value, index, default), tp)



"""Configuration of the broker."""

from __future__ import annotations

import collections.abc
import dataclasses
import fnmatch
import itertools
import logging
import pathlib
import tomllib
import typing

from ..rpclogin import RpcLogin, RpcLoginType
from ..rpcmethod import RpcMethodAccess
from ..rpcri import rpcri_match
from ..rpcurl import RpcUrl
from .configabc import RpcBrokerConfigABC, RpcBrokerRoleABC
from .utils import nmax

logger = logging.getLogger(__name__)


class _NamedProto(typing.Protocol):
    @property
    def name(self) -> str: ...


NamedT = typing.TypeVar("NamedT", bound=_NamedProto)


class NamedMap(collections.abc.Mapping[str, NamedT]):
    """Helper to store and access objects that have name property."""

    def __init__(self, items: collections.abc.Iterable[NamedT] | None = None) -> None:
        self._items = {i.name: i for i in (items or [])}

    def __getitem__(self, name: str) -> NamedT:
        return self._items[name]

    def __iter__(self) -> collections.abc.Iterator[str]:
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NamedMap) and self._items == other._items

    def __repr__(self) -> str:
        return repr(list(self._items.values()))

    def add(self, item: NamedT) -> None:
        """Add a new item. This replaced old user with the same name."""
        self._items[item.name] = item

    def remove(self, user: RpcBrokerConfig.User | str) -> None:
        """Remove item."""
        del self._items[user if isinstance(user, str) else user.name]


@dataclasses.dataclass
class RpcBrokerConfig(RpcBrokerConfigABC):
    """Configuration for the SHV RPC Broker."""

    @dataclasses.dataclass
    class Role:
        """The configured role."""

        name: str
        """Name of this role."""
        mount_points: set[str] = dataclasses.field(default_factory=set)
        """Set of patterns for mount points allowed to this role."""
        access: dict[RpcMethodAccess, set[str]] = dataclasses.field(
            default_factory=dict
        )
        """Resource identifiers used to assign highest possible access level."""

        def access_level(self, path: str, method: str) -> RpcMethodAccess | None:
            """Deduce access level for method based on these rules."""
            for level in sorted(RpcMethodAccess, reverse=True):
                if any(
                    rpcri_match(ri, path, method) for ri in self.access.get(level, [])
                ):
                    return level
            return None

    @dataclasses.dataclass
    class Autosetup:
        """Automatic setup based on device ID for this role."""

        device_id: set[str]
        """Set of patterns for device ID matching."""
        roles: set[str] = dataclasses.field(default_factory=set)
        """Set of roles that must match role assigned to the user."""
        mount_point: str | None = None
        """Mount point assigned by this autosetup.

        Mount point can contain `%` prefix stand-ins that are replaced when
        applied for the real mount point.

        * ``%d`` is replaced with device ID provided by client in login.
        * ``%r`` is replaced with role assigned to the user (multiple roles are
          joined with ``-``).
        * ``%u`` is replaced with user's name.
        * ``%i`` is a unique number. This is used in case mount point without it
          already exists. It starts on ``1`` and the first unused mount point is
          used while counting upward.
        * ``%I`` is a unique number from ``0`` that is increased until not yet
          used mount point is located.
        * ``%%`` is replaced with plain ``%``
        """
        subscriptions: set[str] = dataclasses.field(default_factory=set)
        """Set of initial subscriptions."""

        def generate_mount_point(
            self,
            existing: collections.abc.Set[str],
            device_id: str,
            user: RpcBrokerConfig.User,
        ) -> str | None:
            """Generate an appropriate mount point."""
            if self.mount_point:
                for i in itertools.count():
                    mp = self.mount_point
                    res = ""
                    constlen = None
                    while True:
                        r, sep, mp = mp.partition("%")
                        res += r
                        if sep == "%":
                            match mp[0]:
                                case "d":
                                    res += device_id
                                case "r":
                                    res += "-".join(user.roles)
                                case "u":
                                    res += user.name
                                case "i":
                                    if constlen is None:
                                        constlen = len(res)
                                    if i:
                                        res += str(i)
                                case "I":
                                    if constlen is None:
                                        constlen = len(res)
                                    res += str(i)
                                case "%":
                                    res += "%"
                                case char:
                                    res += "%" + char
                            mp = mp[1:]
                        else:
                            break

                    generate_new = False
                    for mnt in existing:
                        if mnt == res or res.startswith(mnt + "/"):
                            if constlen is None or len(mnt) < constlen:
                                return None  # Can't generate unique path
                            generate_new = True
                    if not generate_new:
                        return res
            return None

    @dataclasses.dataclass
    class User:
        """The user used to login to the broker."""

        name: str
        """Username of this user."""
        password: str
        """Password user needs to use to login."""
        roles: list[str] = dataclasses.field(default_factory=list)
        """Roles assigned to this user."""
        login_type: RpcLoginType = RpcLoginType.PLAIN
        """The password format (the default is SHA1)."""

        class Role(RpcBrokerRoleABC):
            """Broker role for logged in user."""

            def __init__(
                self,
                config: RpcBrokerConfig,
                user: RpcBrokerConfig.User,
                device_id: str | None = None,
                mount_point: str | None = None,
            ) -> None:
                self.config = config
                self.user = user
                self.device_id = device_id
                self._mount_point = mount_point

            @property
            def name(self) -> str:  # noqa D102
                return "-".join(self.user.roles)

            def _autosetup(self) -> RpcBrokerConfig.Autosetup | None:
                if self.device_id is None:
                    return None
                return next(
                    (
                        setup
                        for setup in self.config.autosetups
                        if setup.roles & set(self.user.roles)
                        and any(
                            fnmatch.fnmatchcase(self.device_id, did)
                            for did in setup.device_id
                        )
                    ),
                    None,
                )

            def mount_point(  # noqa D102
                self, existing: collections.abc.Set[str] = frozenset()
            ) -> str | None:
                if self._mount_point is None and (autosetup := self._autosetup()):
                    assert self.device_id is not None
                    return autosetup.generate_mount_point(
                        existing, self.device_id, self.user
                    )
                return self._mount_point

            def initial_subscriptions(self) -> collections.abc.Iterator[str]:  # noqa D102
                if autosetup := self._autosetup():
                    yield from autosetup.subscriptions

            def access_level(  # noqa PLR6301
                self, path: str, method: str
            ) -> RpcMethodAccess | None:
                return nmax(
                    self.config.roles[r].access_level(path, method)
                    for r in self.user.roles
                )

    @dataclasses.dataclass
    class Connect:
        """Connection to be established to some other peer."""

        url: RpcUrl
        """RPC URL specifying the connection destination."""
        roles: list[str] = dataclasses.field(default_factory=list)
        """Roles assigned to this connection."""
        mount_point: str | None = None
        """Mount point for this connection."""
        subscriptions: set[str] = dataclasses.field(default_factory=set)
        """Set of subscriptions to be automatically prepared for this connection."""

        class Role(RpcBrokerRoleABC):
            """The role for the connections."""

            def __init__(
                self, config: RpcBrokerConfig, connection: RpcBrokerConfig.Connect
            ) -> None:
                self.config = config
                self.connection = connection

            @property
            def name(self) -> str:  # noqa D102
                return "+".join(self.connection.roles)

            def mount_point(  # noqa D102
                self, existing: collections.abc.Set[str] = frozenset()
            ) -> str | None:
                return self.connection.mount_point

            def initial_subscriptions(self) -> collections.abc.Iterator[str]:  # noqa D102
                yield from self.connection.subscriptions

            def access_level(  # noqa PLR6301
                self, path: str, method: str
            ) -> RpcMethodAccess | None:
                return nmax(
                    self.config.roles[r].access_level(path, method)
                    for r in self.connection.roles
                )

    def __init__(
        self,
        name: str = "",
        listen: collections.abc.Iterable[RpcUrl] = frozenset(),
        connect: collections.abc.Iterable[RpcBrokerConfig.Connect] = frozenset(),
        roles: collections.abc.Iterable[RpcBrokerConfig.Role] = frozenset(),
        users: collections.abc.Iterable[RpcBrokerConfig.User] = frozenset(),
        autosetups: collections.abc.Iterable[RpcBrokerConfig.Autosetup] = frozenset(),
    ) -> None:
        self._name = name
        self.listen: list[RpcUrl] = list(listen)
        """List of server URLs where broker should listen for peers."""
        self.connect: list[RpcBrokerConfig.Connect] = list(connect)
        """List of connections broker should establish."""
        self.roles: NamedMap[RpcBrokerConfig.Role] = NamedMap(roles)
        """Role descriptions for this configuration."""
        self.users: NamedMap[RpcBrokerConfig.User] = NamedMap(users)
        """Users available in this configuration."""
        self.autosetups: list[RpcBrokerConfig.Autosetup] = list(autosetups)
        """Sequence of autosetup rules."""

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RpcBrokerConfig)
            and self.name == other.name
            and self.listen == other.listen
            and self.connect == other.connect
            and self.roles == other.roles
            and self.users == other.users
            and self.autosetups == other.autosetups
        )

    def __repr__(self) -> str:
        return repr({
            "name": self._name,
            "listen": self.listen,
            "connect": self.connect,
            "roles": self.roles,
            "users": self.users,
            "autosetups": self.autosetups,
        })

    @property
    def name(self) -> str:  # noqa D102
        return self._name

    def listens(self) -> collections.abc.Iterator[RpcUrl]:
        """Iterate over URLs where Broker should listen to."""
        yield from self.listen

    def connections(
        self,
    ) -> collections.abc.Iterator[tuple[RpcUrl, RpcBrokerRoleABC]]:
        """Iterate over URLs and their setup where Broker should connect to."""
        for connect in self.connect:
            yield connect.url, connect.Role(self, connect)

    def login(self, login: RpcLogin, nonce: str) -> RpcBrokerRoleABC | None:  # noqa D102
        if (user := self.users.get(login.username, None)) and login.validate_password(
            user.password, nonce, user.login_type
        ):
            if login.device_mount_point is not None and not any(
                fnmatch.fnmatchcase(login.device_mount_point, dmp)
                for dmp in itertools.chain(
                    *(self.roles[r].mount_points for r in user.roles)
                )
            ):
                raise ValueError("Mount point is not allowed")
            return self.User.Role(self, user, login.device_id, login.device_mount_point)
        return None

    @classmethod
    def load(cls, path: pathlib.Path) -> RpcBrokerConfig:
        """Read configuration from file.

        :param file: Path to the configuration file.
        :return: Broker configuration instance.
        :raise ConfigurationError: When there is an issue in the configuration.
        """
        with path.open("rb") as f:
            data = tomllib.load(f)

        res = cls(str(data.pop("name", "")))
        res.listen = cls._load_urls(data.pop("listen", []), "listen")

        if connects := data.pop("connect", {}):
            if not isinstance(connects, collections.abc.Sequence):
                raise ConfigurationError("'connect' must be array")
            for i, connect in enumerate(connects):
                if not isinstance(connect, collections.abc.MutableMapping):
                    raise ConfigurationError(f"'connect[{i}]' must be table")
                url = connect.pop("url", None)
                if not isinstance(url, str):
                    raise ConfigurationError(
                        f"'connect[{i}].url' must be string and provided"
                    )
                con = cls.Connect(RpcUrl.parse(url))
                con.roles = cls._load_strlist(
                    connect.pop("role", []), f"connect[{i}].role"
                ) or ["default"]
                if (mount_point := connect.pop("mountPoint", None)) is not None:
                    con.mount_point = str(mount_point)
                con.subscriptions = cls._load_ris(
                    connect.pop("subscriptions", []), f"connect[{i}].subscriptions"
                )
                res.connect.append(con)

        if users := data.pop("user", {}):
            if not isinstance(users, collections.abc.Mapping):
                raise ConfigurationError("'user' must be table")
            for name, user in users.items():
                if (password := user.pop("password", None)) is not None:
                    login_type = RpcLoginType.PLAIN
                elif (password := user.pop("sha1pass", None)) is not None:
                    login_type = RpcLoginType.SHA1
                else:
                    raise ConfigurationError(f"'user.{name}.password' must be specfied")
                uroles = cls._load_strlist(
                    user.pop("role", []), f"user.{name}.role"
                ) or ["default"]
                res.users.add(cls.User(name, str(password), uroles, login_type))
                if user:
                    raise ConfigurationError(
                        f"'user.{name}' invalid table keys: {', '.join(user)}"
                    )

        if roles := data.pop("role", {}):
            if not isinstance(roles, collections.abc.Mapping):
                raise ConfigurationError("'role' must be table")
            for name, role in roles.items():
                r = cls.Role(name)
                if (mount_points := role.pop("mountPoints", None)) is not None:
                    if isinstance(mount_points, str):
                        mount_points = [mount_points]
                    if not isinstance(mount_points, collections.abc.Sequence):
                        raise ConfigurationError(
                            f"'role.{name}.mountPoints' must be array of strings"
                        )
                    r.mount_points = {str(v) for v in mount_points}
                if (access := role.pop("access", None)) is not None:
                    if not isinstance(access, collections.abc.Mapping):
                        raise ConfigurationError(f"'role.{name}.access' must be table")
                    for level, paths in access.items():
                        nlevel = RpcMethodAccess.strmap().get(level)
                        if nlevel is None:
                            raise ConfigurationError(
                                f"'{level} is not allowed in 'role.{name}.access'"
                            )
                        if not isinstance(paths, collections.abc.Sequence):
                            raise ConfigurationError(
                                f"'role.{name}.access.{level}' must be array"
                            )
                        r.access[nlevel] = {
                            str(v)
                            for v in ([paths] if isinstance(paths, str) else paths)
                        }
                if role:
                    raise ConfigurationError(
                        f"'role.{name}' invalid table keys: {', '.join(role)}"
                    )
                res.roles.add(r)

        if autosetups := data.pop("autosetup", []):
            if not isinstance(autosetups, collections.abc.Sequence):
                raise ConfigurationError("'role' must be array")
            for i, autosetup in enumerate(autosetups):
                if not isinstance(autosetup, collections.abc.MutableMapping):
                    raise ConfigurationError(f"'autosetup[{i}]' must be table")
                ast = cls.Autosetup(
                    cls._load_strset(
                        autosetup.pop("deviceId", []), f"autosetup[{i}].deviceId"
                    )
                )
                ast.roles = cls._load_strset(
                    autosetup.pop("role", []), f"autosetup[{i}].role"
                )
                if (mount_point := autosetup.pop("mountPoint", None)) is not None:
                    ast.mount_point = str(mount_point)
                ast.subscriptions = cls._load_ris(
                    autosetup.pop("subscriptions", []), f"autosetup[{i}].subscriptions"
                )
                res.autosetups.append(ast)

        if data:
            raise ConfigurationError(f"Invalid table keys: {', '.join(data)}")

        return res

    @staticmethod
    def _load_strarr(value: object, location: str) -> collections.abc.Iterator[str]:
        if isinstance(value, str):
            value = [value]
        if not isinstance(value, collections.abc.Sequence) or any(
            not isinstance(v, str) for v in value
        ):
            raise ConfigurationError(f"'{location}' must be array of strings")
        yield from value

    @classmethod
    def _load_strlist(cls, value: object, location: str) -> list[str]:
        return list(cls._load_strarr(value, location))

    @classmethod
    def _load_strset(cls, value: object, location: str) -> set[str]:
        return set(cls._load_strarr(value, location))

    @classmethod
    def _load_urls(cls, value: object, location: str) -> list[RpcUrl]:
        return [RpcUrl.parse(sub) for sub in cls._load_strarr(value, location)]

    @classmethod
    def _load_ris(cls, value: object, location: str) -> set[str]:
        # TODO possibly validate RIs
        return {sub for sub in cls._load_strarr(value, location)}


class ConfigurationError(ValueError):
    """The error in the configuration."""



"""The base for the various high level SHV RPC interfaces."""

import asyncio
import collections.abc
import contextlib
import datetime
import logging
import traceback

from .__version__ import VERSION
from .rpcerrors import (
    RpcError,
    RpcInvalidParamError,
    RpcMethodCallExceptionError,
    RpcMethodNotFoundError,
    RpcUserIDRequiredError,
)
from .rpcmessage import RpcMessage
from .rpcmethod import RpcMethodAccess, RpcMethodDesc
from .rpctransport import RpcClient
from .shvversion import SHV_VERSION_MAJOR, SHV_VERSION_MINOR
from .value import SHVType, is_shvbool, is_shvnull

logger = logging.getLogger(__name__)


class SimpleBase:
    """SHV RPC made simple to use.

    You want to use this if you plan to implement your own specific RPC handler
    but in most cases you might want to use :class:`SimpleClient` or
    :class:`SimpleDevice` instead.

    Messages are handled in an asyncio loop and based on the type of the message
    the different operation is performed.
    """

    IDLE_TIMEOUT: float = 180
    """Number of seconds before we are disconnected from the broker automatically."""

    APP_NAME: str = "pyshv"
    """Name of the application reported to the SHV.

    You can change this value in child class to report a more accurate
    application name.
    """
    APP_VERSION: str = VERSION
    """Version of the application reported to the SHV.

    You should change this value in child class to report a correct number.

    :param client: The RPC client instance to wrap and manage.
    :param call_attempts: Number of attempts for :meth:`call` when no response
      is received before call is abandoned. You can use zero (or negative
      number) for unlimited number of attempts. The default is a single attempt
      and this only regular one call.
    :param call_timeout: Timeout in seconds before call is attempted again or
      abandoned (if there was too much call attempts). This is time before we
      consider response to be lost.
    """

    def __init__(
        self,
        client: RpcClient,
        call_attempts: int = 1,
        call_timeout: float | None = 300.0,
    ) -> None:
        self.client = client
        """The underlaying RPC client instance.

        **Do not send messages by directly accessing this property. You must use
        implementations provided by this class!**
        """
        self.task = asyncio.create_task(self._loop())
        """Task running the message handling receive loop."""
        self.call_attempts = call_attempts
        """Number of attempts when no response is received before call is abandoned.
        You can use zero for no  or negative number to have unlimited attempts.
        """
        self.call_timeout = call_timeout
        """Timeout in seconds before call is attempted again or abandoned."""
        self._calls_event: dict[int, asyncio.Event] = {}
        self._calls_msg: dict[int, RpcMessage] = {}
        self.__peer_is_shv3: None | bool = None

    async def disconnect(self) -> None:
        """Disconnect an existing connection.

        The call to the disconnect when client is not connected is silently
        ignored.

        Await blocks until disconnect is completed. You can use
        ``x.client.disconnect()`` to only initialize disconnect without waiting
        for it to take an effect.
        """
        self.client.disconnect()
        await self.task

    async def _loop(self) -> None:
        """Loop run in asyncio task to receive messages."""
        async with asyncio.TaskGroup() as tg:
            with contextlib.suppress(EOFError):
                while msg := await self.client.receive(raise_error=False):
                    if msg is RpcClient.Control.RESET:
                        self._reset()
                    elif msg.is_valid():
                        tg.create_task(self._message(msg))
                    else:
                        logger.info("%s: Dropped invalid message: %s", self.client, msg)

    def _reset(self) -> None:
        """Handle peer's reset request."""
        logger.info("%s: Doing reset", self.client)
        self.__peer_is_shv3 = None
        for event in self._calls_event.values():
            event.set()

    async def reset(self) -> None:
        """Reset the client and connection.

        This calls :meth:`RpcClient.reset` as well as reset of the internal
        state.
        """
        await self.client.reset()
        self._reset()

    async def _send(self, msg: RpcMessage) -> None:
        """Send message.

        Use this only if you want send a generic message (such as when you are
        passing message along). :meth:`call` or :meth:`_signal` should be
        prefered when ever possible.

        You should be using this method instead of ``self.client.send`` to
        ensure that send can be correctly overwritten and optionally postponed
        or blocked by child implementations.
        """
        # TODO possibly lock to prevent from spliting this call
        await self.client.send(msg)

    async def call(
        self,
        path: str,
        method: str,
        param: SHVType = None,
        call_attempts: int | None = None,
        call_timeout: float | None = None,
        user_id: str | None = None,
    ) -> SHVType:
        """Call given method on given path with given parameter.

        Note that this is coroutine and thus it is up to you if you await it or
        use asyncio tasks.

        The delivery of the messages is not ensure in SHV network (they can be
        dropped due to multiple reasons without informing the source of the
        message) and thus this method can send the request message multiple
        times if it doesn't receive an appropriate response.

        :param path: SHV path method is associated with.
        :param method: SHV method name to be called.
        :param param: Parameter passed to the called method.
        :param call_attempts: Allows override of the object setting.
        :param call_timeout: Allows override of the object setting.
        :param user_id: UserID added to the method call request. This is required
          by some RPC methods to identify the user and they will respond with
          :class:`RpcUserIDRequiredError` if it is missing. This is caught by
          this method and request will be attempted again (not counting in
          ``call_attempts``) with User ID ``""``. If you know that method needs
          User ID then you can prevent this round trip by setting this argument
          to ``""``. On the other hand sending all requests with User ID wastes
          with bandwidth.
        :return: Return value on successful method call.
        :raise RpcError: The call result in error that is propagated by raising
            `RpcError` or its children based on the failure.
        :raise TimeoutError: when response is not received before timeout with
            all attempts depleted.
        :raise EOFError: when client disconnected and thus request can't be sent
          or response received.
        """
        call_attempts = self.call_attempts if call_attempts is None else call_attempts
        call_timeout = self.call_timeout if call_timeout is None else call_timeout
        request = RpcMessage.request(path, method, param, user_id=user_id)
        event = asyncio.Event()
        self._calls_event[request.request_id] = event
        attempt = 0
        while call_attempts < 1 or attempt < call_attempts:
            attempt += 1
            await self._send(request)
            try:
                async with asyncio.timeout(call_timeout):
                    await event.wait()
            except TimeoutError:
                continue
            if request.request_id not in self._calls_msg:
                self._calls_event[request.request_id].clear()
                continue
            response = self._calls_msg.pop(request.request_id)
            if response.is_error:
                rpc_error = response.rpc_error
                if not isinstance(rpc_error, RpcUserIDRequiredError):
                    raise response.rpc_error
                attempt -= 1  # Annul this attempt
                request.user_id = ""
                event.clear()
                self._calls_event[request.new_request_id()] = event
                continue
            return response.result
        raise TimeoutError

    async def ping(self) -> None:
        """Ping the peer to check the connection."""
        await self.call(".app" if await self.peer_is_shv3() else ".broker/app", "ping")

    async def ls(self, path: str) -> list[str]:
        """List child nodes of the node on the specified path.

        :param path: SHV path to the node we want children to be listed for.
        :return: list of child nodes.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "ls")
        if not isinstance(res, list):  # pragma: no cover
            raise RpcMethodCallExceptionError(f"Invalid result returned: {res!r}")
        return res

    async def ls_has_child(self, path: str, name: str) -> bool:
        """Use ``ls`` method to check for child.

        :param path: SHV path to the node with possible child of given name.
        :param name: Name of the child node.
        :return: ``True`` if there is such child and ``False`` if not.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "ls", name)
        if not isinstance(res, bool):  # pragma: no cover
            raise RpcMethodCallExceptionError(f"Invalid result returned: {res!r}")
        return res

    async def dir(self, path: str, details: bool = False) -> list[RpcMethodDesc]:
        """List methods associated with node on the specified path.

        :param path: SHV path to the node we want methods to be listed for.
        :param details: If detailed listing should be performed instead of standard one.
        :return: list of the node's methods.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "dir", True if details else None)
        if isinstance(res, list):  # pragma: no cover
            return [RpcMethodDesc.from_shv(m) for m in res]
        raise RpcMethodCallExceptionError(f"Invalid result returned: {res!r}")

    async def dir_exists(self, path: str, name: str) -> bool:
        """Check if method exists using ``dir`` method.

        :param path: SHV path to the node we want methods to be listed for.
        :param name: Name of the method which existence should be checked.
        :return: ``True`` if method exists and ``False`` otherwise.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "dir", name)
        # list, null and mapping is backward compatibility
        if not isinstance(
            res, bool | None | collections.abc.Mapping | collections.abc.Sequence
        ):  # pragma: no cover
            raise RpcMethodCallExceptionError(f"Invalid result returned: {res!r}")
        return bool(res)

    async def peer_is_shv3(self) -> bool:
        """Check if peer supports at least SHV 3.0."""
        if self.__peer_is_shv3 is None:
            try:
                major = await self.call(".app", "shvVersionMajor")
                self.__peer_is_shv3 = isinstance(major, int) and major >= 0
            except RpcError:
                self.__peer_is_shv3 = False
        return self.__peer_is_shv3

    async def _signal(
        self,
        path: str,
        name: str = "chng",
        source: str = "get",
        value: SHVType = None,
        access: RpcMethodAccess = RpcMethodAccess.READ,
    ) -> None:
        """Send signal from given path and method source and with given parameter.

        Note that this is coroutine and thus it is up to you if you await it or
        use asyncio tasks.

        This is intentionally marked as accessible only by class methods because
        signals must be raised only for valid paths and methods and that is
        something only class itself can ensure. Your implemntation should
        provide public methods that protects against call with invalid path or
        method.

        :param path: SHV path signal is associated with.
        :param name: Signal name to be raised.
        :param source: Method name this signal is associated with.
        :param value: Parameter that is the signaled value.
        :param access: Minimal access level needed to access the signal.
        """
        await self._send(RpcMessage.signal(path, name, source, value, access))

    async def _message(self, msg: RpcMessage) -> None:
        """Handle every received message."""
        if msg.is_request:
            resp = msg.make_response()
            method = msg.method
            try:
                resp.result = await self._method_call(
                    msg.path,
                    method,
                    msg.param,
                    msg.rpc_access or RpcMethodAccess.BROWSE,
                    msg.user_id,
                )
            except RpcError as exp:
                resp.rpc_error = exp
            except Exception as exc:
                resp.rpc_error = RpcMethodCallExceptionError(
                    "".join(traceback.format_exception(exc))
                )
            try:
                await self._send(resp)
            except EOFError:
                return  # No need to spam logs on disconnect
            except Exception as exc:
                logger.warning("%s: Failed to send response", self.client, exc_info=exc)
        elif msg.is_response:
            rid = msg.request_id
            if rid in self._calls_event:
                self._calls_msg[rid] = msg
                self._calls_event.pop(rid).set()
        elif msg.is_signal:
            await self._got_signal(msg.path, msg.signal_name, msg.source, msg.param)

    async def _method_call(
        self,
        path: str,
        method: str,
        param: SHVType,
        access: RpcMethodAccess,
        user_id: str | None,
    ) -> SHVType:
        """Handle request.

        :param path: SHV path to the node the method is associated with.
        :param method: method requested to be called.
        :param param: Parameter to be passed to the called method.
        :param access: access level of the client specified in the request.
        :param client_id: The client's ID collected as message was passed
          around. This can be ``None`` when request message contained no ID. You
          can raise :class:`RpcUserIDRequiredError` if you need it.
        :return: result of the method call. To report error you should raise
            :exc:`RpcError`.
        """
        match path, method:
            case _, "ls":
                return self._method_call_ls(path, param)
            case _, "dir":
                return self._method_call_dir(path, param)
            case ".app", "shvVersionMajor":
                return SHV_VERSION_MAJOR
            case ".app", "shvVersionMinor":
                return SHV_VERSION_MINOR
            case ".app", "name":
                return self.APP_NAME
            case ".app", "version":
                return self.APP_VERSION
            case ".app", "date":
                return datetime.datetime.now().astimezone()
            case ".app", "ping":
                return None
        raise RpcMethodNotFoundError(
            f"No such path '{path}' or method '{method}' or access rights."
        )

    def _method_call_ls(self, path: str, param: SHVType) -> SHVType:
        """Implement ``ls`` method call functionality."""
        if is_shvnull(param):
            res = []
            for v in self._ls(path):
                if v not in res:
                    res.append(v)
            if not res and not self._valid_path(path):
                raise RpcMethodNotFoundError(f"No such node: {path}")
            return res
        if isinstance(param, str):
            return any(v == param for v in self._ls(path))
        raise RpcInvalidParamError("Use Null or String with node name")

    def _ls(self, path: str) -> collections.abc.Iterator[str]:  # noqa: PLR6301
        """Implement ``ls`` method for all nodes.

        The default implementation supports `.app` path. Your implementation
        should yield child nodes.

        Always call this as first before you yield your own methods to provide
        users with standard nodes.

        :param path: SHV path that should be listed.
        :return: Iterator over child nodes of the node on the path.
        """
        if not path:
            yield ".app"

    def _valid_path(self, path: str) -> bool:
        """Check that :meth:`_ls` reports this path as existing.

        :param path: SHV path to be validated.
        :return: ``True`` if :meth:`_ls` on parent node reports node with this
          tail piece. ``False`` is returned otherwise.
        """
        if not path:
            return True  # The root path always exists
        root, _, name = path.rpartition("/")
        return any(name == v for v in self._ls(root))

    def _method_call_dir(self, path: str, param: SHVType) -> SHVType:
        """Implement ``dir`` method call functionality."""
        if not self._valid_path(path):
            raise RpcMethodNotFoundError(f"No such node: {path}")
        if is_shvnull(param) or is_shvbool(param):
            return list(d.to_shv(bool(param)) for d in self._dir(path))
        if isinstance(param, str):
            return any(v.name == param for v in self._dir(path))
        raise RpcInvalidParamError("Use Null or Bool or String with node name")

    def _dir(self, path: str) -> collections.abc.Iterator[RpcMethodDesc]:  # noqa: PLR6301
        """Implement ``dir`` method for all nodes.

        This implementation is called only for valid paths (:meth:`_valid_path`).

        Always call this as first before you yield your own methods to provide
        users with standard ones.

        :param path: SHV path method should be listed for.
        :return: List of methods associated with given node.
        """
        yield RpcMethodDesc.stddir()
        yield RpcMethodDesc.stdls()
        if path == ".app":
            yield RpcMethodDesc.getter("shvVersionMajor", "Null", "Int")
            yield RpcMethodDesc.getter("shvVersionMinor", "Null", "Int")
            yield RpcMethodDesc.getter("name", "Null", "String")
            yield RpcMethodDesc.getter("version", "Null", "String")
            yield RpcMethodDesc.getter("date", "Null", "DateTime")
            yield RpcMethodDesc("ping")

    async def _got_signal(
        self, path: str, signal: str, source: str, value: SHVType
    ) -> None:
        """Handle signal.

        :param path: SHV path to the node the signal is associated with.
        :param signal: Signal name.
        :param source: Method name signal is associated with.
        :param value: The value caried by signal.
        """
        if signal.endswith("chng") and source == "get":
            await self._value_update(path, value)

    async def _value_update(self, path: str, value: SHVType) -> None:
        """Handle value change (``*chng`` signal associated with ``get`` method)."""



"""Check various clients and servers implementing a different link layers."""

from __future__ import annotations

import asyncio
import collections.abc
import contextlib
import dataclasses
import io
import logging
import os
import pty
import select
import threading

import pytest

from shv import (
    RpcClient,
    RpcClientPipe,
    RpcClientTCP,
    RpcClientTTY,
    RpcClientUnix,
    RpcClientWebSockets,
    RpcInvalidRequestError,
    RpcMessage,
    RpcServerTCP,
    RpcServerTTY,
    RpcServerUnix,
    RpcServerWebSockets,
    RpcServerWebSocketsUnix,
)

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class PTYPort:
    """PTY based exchange port."""

    port1: str
    port2: str
    event: threading.Event

    @classmethod
    @contextlib.contextmanager
    def new(cls) -> collections.abc.Iterator[PTYPort]:
        """Context function that provides PTY based ports.

        :return: Tuple with two strings containing paths to the pty ports and
          event that can be cleared to pause data exchange and set to resume it
          again.
        """
        pty1_master, pty1_slave = pty.openpty()
        pty2_master, pty2_slave = pty.openpty()
        event = threading.Event()
        event.set()
        thread = threading.Thread(
            target=cls._exchange, args=(pty1_master, pty2_master, event)
        )
        thread.start()
        yield cls(os.ttyname(pty1_slave), os.ttyname(pty2_slave), event)
        os.close(pty1_slave)
        os.close(pty2_slave)
        thread.join()

    @staticmethod
    def _exchange(fd1: int, fd2: int, event: threading.Event) -> None:
        p = select.poll()
        p.register(fd1, select.POLLIN | select.POLLPRI)
        p.register(fd2, select.POLLIN | select.POLLPRI)
        fds = 2
        while fds:
            for fd, pevent in p.poll():
                event.wait()
                if pevent & select.POLLHUP:
                    p.unregister(fd)
                    fds -= 1
                if pevent & select.POLLIN:
                    data = os.read(fd, io.DEFAULT_BUFFER_SIZE)
                    os.write(fd1 if fd is fd2 else fd2, data)
                if pevent & select.POLLERR:
                    logger.error("Error detected in ptycopy")
                    return


@pytest.fixture(name="pty")
def fixture_pty():
    with PTYPort.new() as port:
        yield port


class Link:
    """Generic tests applied to link layers."""

    async def test_call(self, clients):
        msg = RpcMessage.request(".app", "dir")
        await clients[0].send(msg)
        assert await clients[1].receive() == msg

    async def test_notify(self, clients):
        msg = RpcMessage.signal(".app")
        await clients[1].send(msg)
        assert await clients[0].receive() == msg

    async def test_error_receive(self, clients):
        msg = RpcMessage.request(".app", "dir").make_response()
        msg.rpc_error = RpcInvalidRequestError("Fake error")
        await clients[0].send(msg)
        assert await clients[1].receive(False) == msg

    async def test_error_raise(self, clients):
        msg = RpcMessage.request(".app", "dir").make_response()
        msg.rpc_error = RpcInvalidRequestError("Fake error")
        await clients[0].send(msg)
        with pytest.raises(RpcInvalidRequestError):
            await clients[1].receive(True)

    @pytest.mark.parametrize("a,b", ((0, 1), (1, 0)))
    async def test_reset(self, clients, a, b):
        await clients[a].reset()
        assert await clients[b].receive() == RpcClient.Control.RESET


class ServerLink(Link):
    """Additional tests for server-client connections."""

    @pytest.mark.parametrize("cdisc,crecv", ((0, 1), (1, 0)))
    async def test_eof_receive(self, clients, cdisc, crecv):
        clients[cdisc].disconnect()
        await clients[cdisc].wait_disconnect()
        assert not clients[cdisc].connected
        with pytest.raises(EOFError):
            await clients[crecv].receive()
        assert not clients[crecv].connected

    @pytest.mark.parametrize("cdisc,crecv", ((0, 1), (1, 0)))
    async def test_eof_send(self, clients, cdisc, crecv):
        clients[cdisc].disconnect()
        await clients[cdisc].wait_disconnect()
        with pytest.raises(EOFError):
            await clients[crecv].send(RpcMessage.request(".app", "ping"))
        assert not clients[crecv].connected

    async def test_reconnect(self, clients, server):
        clients[1].disconnect()
        await clients[1].wait_disconnect()
        await clients[1].reset()
        server_client = await server[1].get()
        msg = RpcMessage.request("foo", "ls")
        await clients[1].send(msg)
        assert await server_client.receive() == msg
        server_client.disconnect()
        await server_client.wait_disconnect()


class TestPipe(Link):
    """Check that we can work over Unix pipe pair."""

    @pytest.fixture(name="clients")
    async def fixture_clients(self):
        client1, client2 = await RpcClientPipe.open_pair()
        yield client1, client2
        client1.disconnect()
        client2.disconnect()
        await client1.wait_disconnect()
        await client2.wait_disconnect()


class TestTCP(ServerLink):
    """Check that TCP/IP transport protocol works."""

    @pytest.fixture(name="server")
    async def fixture_server(self, port):
        queue = asyncio.Queue()
        server = RpcServerTCP(queue.put, "localhost", port)
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, port):
        client = await RpcClientTCP.connect("localhost", port)
        server_client = await server[1].get()
        yield server_client, client
        client.disconnect()
        server_client.disconnect()
        await client.wait_disconnect()
        await server_client.wait_disconnect()

    async def test_before_connect(self, port) -> None:
        client = RpcClientTCP("localhost", port)
        with pytest.raises(EOFError):
            await client.receive()
        with pytest.raises(EOFError):
            await client.send(RpcMessage.request(".app", "ping"))
        client.disconnect()
        await client.wait_disconnect()


class TestUnix(ServerLink):
    """Check that Unix transport protocol works."""

    @pytest.fixture(name="sockpath")
    def fixture_sockpath(self, tmp_path):
        return tmp_path / "s"

    @pytest.fixture(name="server")
    async def fixture_server(self, sockpath):
        queue = asyncio.Queue()
        server = RpcServerUnix(queue.put, sockpath)
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, sockpath):
        client = await RpcClientUnix.connect(sockpath)
        server_client = await server[1].get()
        yield server_client, client
        server_client.disconnect()
        client.disconnect()
        await server_client.wait_disconnect()
        await client.wait_disconnect()


class TestTTY(Link):
    """Check that we can work over TTY port transport protocol."""

    @pytest.fixture(name="clients")
    async def fixture_clients(self, pty):
        # We pause data exchange to resolve race condition between terminal
        # setup and reset message sending. This ensures that ports are
        # configured to raw mode before reset message is exchanged.
        pty.event.clear()
        client1 = await RpcClientTTY.connect(pty.port1)
        client2 = await RpcClientTTY.connect(pty.port2)
        pty.event.set()

        # Flush reset sent by clients
        assert await client1.receive() is RpcClient.Control.RESET
        assert await client2.receive() is RpcClient.Control.RESET

        yield client1, client2

        client1.disconnect()
        client2.disconnect()
        await client1.wait_disconnect()
        await client2.wait_disconnect()

    async def test_escapes(self, clients):
        msg = RpcMessage.request("prop", "set", b"1\xa2\xa3\xa4\xa5\xaa2")
        await clients[0].send(msg)
        assert await clients[1].receive() == msg


class TestTTYServer(Link):
    """Check that we can work over TTY port transport protocol and its server."""

    @pytest.fixture(name="server")
    async def fixture_server(self, pty):
        pty.event.clear()
        queue = asyncio.Queue()
        server = RpcServerTTY(queue.put, pty.port1)
        await server.listen()
        yield server, queue
        server.terminate()
        await server.wait_terminated()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, pty):
        client = await RpcClientTTY.connect(pty.port2)
        server_client = await server[1].get()
        pty.event.set()
        assert await server_client.receive() is RpcClient.Control.RESET
        assert await client.receive() is RpcClient.Control.RESET
        yield server_client, client
        client.disconnect()
        server_client.disconnect()
        await client.wait_disconnect()
        await server_client.wait_disconnect()


class TestWebSockets(ServerLink):
    """Check that WebSockets transport protocol works."""

    @pytest.fixture(name="server")
    async def fixture_server(self, port):
        queue = asyncio.Queue()
        server = RpcServerWebSockets(queue.put, "localhost", port)
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, port):
        client = await RpcClientWebSockets.connect("localhost", port)
        server_client = await server[1].get()
        await asyncio.sleep(0)
        yield server_client, client
        client.disconnect()
        server_client.disconnect()
        await client.wait_disconnect()
        await server_client.wait_disconnect()

    async def test_before_connect(self, port) -> None:
        client = RpcClientWebSockets("localhost", port)
        with pytest.raises(EOFError):
            await client.receive()
        with pytest.raises(EOFError):
            await client.send(RpcMessage.request(".app", "ping"))
        client.disconnect()
        await client.wait_disconnect()


class TestWebSocketsUnix(ServerLink):
    """Check that WebSockets Unix transport protocol works."""

    @pytest.fixture(name="sockpath")
    def fixture_sockpath(self, tmp_path):
        return tmp_path / "s"

    @pytest.fixture(name="server")
    async def fixture_server(self, sockpath):
        queue = asyncio.Queue()
        server = RpcServerWebSocketsUnix(queue.put, sockpath)
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, sockpath):
        client = await RpcClientWebSockets.connect(sockpath)
        server_client = await server[1].get()
        yield server_client, client
        server_client.disconnect()
        client.disconnect()
        await server_client.wait_disconnect()
        await client.wait_disconnect()



"""Resource identification matching in SHV RPC.

This is implemented according to the `SHV standard documentation
<https://silicon-heaven.github.io/shv-doc/rpcri.html>`__.
"""

from __future__ import annotations

import fnmatch

from .value import SHVType


def rpcri_match(ri: str, path: str, method: str, signal: str | None = None) -> bool:
    """Check if given path or method matches given RPC RI.

    This can use used to match method or signal against RI.

    :param ri: RPC RI that should be used to match method or signal.
    :param path: SHV Path that that RI should match.
    :param method: SHV RPC method name that RI should match.
    :param signal: SHV RPC signal name that RI should match. This can be
      ``NULL`` and in such case RI matches method (``PATH:METHOD``) or non-empty
      string for signal matching (``PATH:METHOD:SIGNAL``).
    :return: ``True`` if RI matches the provided method or signal and ``False``
      otherwise.
    """
    parts = ri.split(":")
    match len(parts):
        case 2:
            return (
                signal is None
                and shvpath_match(parts[0], path)
                and fnmatch.fnmatchcase(method, parts[1])
            )
        case 3:
            return (
                signal is not None
                and shvpath_match(parts[0], path)
                and fnmatch.fnmatchcase(method, parts[1])
                and fnmatch.fnmatchcase(signal, parts[2])
            )
    return False


def rpcri_relative_to(ri: str, path: str) -> str | None:
    """Derive the RPC RI that is relative to the given path.

    :param ri: RPC RI that should be modified to be relative to the given path.
    :param path: Path that must be used as a relative root.
    :return: New RPC RI or ``None`` in case RI is not relative to the path.
    """
    pth, sep, rest = ri.partition(":")
    npth = shvpath_tail(pth, path)
    return f"{npth}{sep}{rest}" if npth is not None else None


def rpcri_legacy_subscription(ri: str) -> SHVType:
    """Convert RPC RI to legacy SHV subscription representation.

    :param ri: RPC RI for signal matching that should be converted.
    :return: Legacy subscription description.
    """
    parts = ri.split(":")
    if len(parts) != 3:
        raise ValueError("Must be RPC RI for signals")
    path, method, signal = parts
    res: dict[str, SHVType] = {}
    pth, _, tail = path.rpartition("/")
    if "*" in pth or tail != "**":
        res["paths"] = path
    else:
        res["path"] = pth
    if "*" in signal and signal != "*":
        res["methods"] = signal
    else:
        res["method"] = "" if signal == "*" else signal
    if method != "*":
        res["source"] = method
    return res


def __pth_match(path: str, pattern: list[str]) -> int | None:
    i = 0
    for node in path.split("/"):
        if i >= len(pattern):
            return None
        if pattern[i] == "**":
            if len(pattern) == i + 1:
                return i + 1  # Matches everything so just return
            if pattern[i + 1] == "**":
                i += 1
            elif fnmatch.fnmatchcase(node, pattern[i + 1]):
                i += 2
            continue
        if not fnmatch.fnmatchcase(node, pattern[i]):
            return None
        i += 1
    return i


def shvpath_match(pattern: str, path: str) -> bool:
    """Check if given path matches given pattern.

    :param pattern: Pattern that should match the SHV path.
    :param path: SHV Path.
    :return: ``True`` if the whole path matches the pattern and ``False`` otherwise.
    """
    ptn = pattern.split("/")
    res = __pth_match(path, ptn)
    return res is not None and len(ptn) == res


def shvpath_tail(pattern: str, path: str) -> str | None:
    """Remove the pattern prefix that matches given path.

    :param pattern: Pattern for which tail should be derived.
    :param path: Path the pattern should match.
    :return: Returns tail that can be used to match nodes bellow path or ``None`` in
      case pattern doesn't match the path or matches it exactly.
    """
    ptn = pattern.split("/")
    res = __pth_match(path, ptn)
    if len(ptn) == res and ptn[-1] == "**":
        res -= 1
    if res is None or len(ptn) == res:
        return None
    return "/".join(ptn[res:])



"""Common base for stream based clients."""

from __future__ import annotations

import abc
import asyncio
import binascii
import collections.abc
import contextlib
import logging
import typing
import weakref

from ..chainpack import ChainPack
from .abc import RpcClient, RpcServer

logger = logging.getLogger(__name__)


class RpcTransportProtocol(abc.ABC):
    """Base for the implementations of RPC transport protocols."""

    @classmethod
    @abc.abstractmethod
    async def send(
        cls,
        write: collections.abc.Callable[[bytes], collections.abc.Awaitable[None]],
        msg: bytes,
    ) -> None:
        """Send message.

        :param msg: Bytes of a complete message to be sent.
        """

    @classmethod
    @abc.abstractmethod
    async def receive(
        cls,
        read: collections.abc.Callable[[int], collections.abc.Awaitable[bytes]],
    ) -> bytes:
        """Receive message.

        :return: Bytes of complete message.
        :raise EOFError: when EOF is encountered.
        """

    @classmethod
    async def asyncio_send(cls, writer: asyncio.StreamWriter, msg: bytes) -> None:
        """Variation on :meth:`send` that uses :class:`asyncio.StreamWriter`."""

        async def write(d: bytes) -> None:
            writer.write(d)
            await writer.drain()

        await cls.send(write, msg)

    @classmethod
    async def asyncio_receive(cls, reader: asyncio.StreamReader) -> bytes:
        """Variation on :meth:`receive` that uses :class:`asyncio.StreamReader`."""

        async def read(n: int) -> bytes:
            try:
                return await reader.readexactly(n)
            except (asyncio.IncompleteReadError, ConnectionError) as exc:
                raise EOFError from exc

        return await cls.receive(read)


class RpcProtocolStream(RpcTransportProtocol):
    """SHV RPC Stream protocol."""

    @classmethod
    async def send(  # noqa: D102
        cls,
        write: collections.abc.Callable[[bytes], collections.abc.Awaitable[None]],
        msg: bytes,
    ) -> None:
        await write(ChainPack.pack_uint_data(len(msg)))
        await write(msg)

    @classmethod
    async def receive(  # noqa: D102
        cls,
        read: collections.abc.Callable[[int], collections.abc.Awaitable[bytes]],
    ) -> bytes:
        sdata = bytearray()
        while True:
            sdata += await read(1)
            try:
                size = ChainPack.unpack_uint_data(sdata)
            except ValueError:
                pass
            else:
                return await read(size)


class _RpcProtocolSerial(RpcTransportProtocol):
    """SHV RPC Serial protocol."""

    STX = 0xA2
    ETX = 0xA3
    ATX = 0xA4
    ESC = 0xAA
    ESCMAP: typing.Final = {0x02: STX, 0x03: ETX, 0x04: ATX, 0x0A: ESC}
    ESCRMAP: typing.Final = {v: k for k, v in ESCMAP.items()}

    @classmethod
    async def _send(
        cls,
        write: collections.abc.Callable[[bytes], collections.abc.Awaitable[None]],
        msg: bytes,
        use_crc: bool,
    ) -> None:
        await write(bytes((cls.STX,)))
        escmsg = cls.escape(msg)
        await write(escmsg)
        await write(bytes((cls.ETX,)))
        if use_crc:
            await write(cls.escape(binascii.crc32(escmsg).to_bytes(4, "big")))

    @classmethod
    def escape(cls, data: bytes) -> bytes:
        """Escape bytes as defined for serial protocol.

        :param data: Data to be escaped (node that you can pass
          :class:`bytearray` to modify in place).
        :return: The modified data.
        """
        # Warning: ESC must be first in the list so we first replace all occurances of
        # ESC before we insert more of them.
        for b in (cls.ESC, cls.STX, cls.ETX, cls.ATX):
            data = data.replace(bytes((b,)), bytes((cls.ESC, cls.ESCRMAP[b])))
        return data

    @classmethod
    async def _receive(
        cls,
        read: collections.abc.Callable[[int], collections.abc.Awaitable[bytes]],
        use_crc: bool,
    ) -> bytes:
        while True:
            while (await read(1))[0] != cls.STX:
                pass
            data = bytearray()
            while (b := (await read(1))[0]) not in {cls.ETX, cls.ATX}:
                data += bytes((b,))
            if b == cls.ATX:
                continue
            if use_crc:
                crc32_br = bytearray()
                while len(crc32_br) < (siz := 4 + crc32_br.count(bytes((cls.ESC,)))):
                    crc32_br += await read(siz - len(crc32_br))
                crc32_b = cls.deescape(crc32_br)
                if int.from_bytes(crc32_b, "big") != binascii.crc32(data):
                    continue
            return cls.deescape(data)

    @classmethod
    def deescape(cls, data: bytes) -> bytes:
        """Reverse escape operation on bytes as defined for serial protocol.

        :param data: Escaped data (node that you can pass :class:`bytearray` to
          modify in place).
        :return: The modified data.
        """
        for b in (cls.STX, cls.ETX, cls.ATX, cls.ESC):
            data = data.replace(bytes((cls.ESC, cls.ESCRMAP[b])), bytes((b,)))
        return data


class RpcProtocolSerial(_RpcProtocolSerial):
    """SHV RPC Serial protocol."""

    @classmethod
    async def send(  # noqa: D102
        cls,
        write: collections.abc.Callable[[bytes], collections.abc.Awaitable[None]],
        msg: bytes,
    ) -> None:
        await cls._send(write, msg, False)

    @classmethod
    async def receive(  # noqa: D102
        cls,
        read: collections.abc.Callable[[int], collections.abc.Awaitable[bytes]],
    ) -> bytes:
        return await cls._receive(read, False)


class RpcProtocolSerialCRC(_RpcProtocolSerial):
    """SHV RPC Serial protocol with CRC32 message validation."""

    @classmethod
    async def send(  # noqa: D102
        cls,
        write: collections.abc.Callable[[bytes], collections.abc.Awaitable[None]],
        msg: bytes,
    ) -> None:
        await cls._send(write, msg, True)

    @classmethod
    async def receive(  # noqa: D102
        cls,
        read: collections.abc.Callable[[int], collections.abc.Awaitable[bytes]],
    ) -> bytes:
        return await cls._receive(read, True)


class RpcClientStream(RpcClient):
    """RPC connection to some SHV peer over data stream."""

    def __init__(
        self,
        protocol: type[RpcTransportProtocol] = RpcProtocolStream,
    ) -> None:
        super().__init__()
        self._reader: None | asyncio.StreamReader = None
        self._writer: None | asyncio.StreamWriter = None
        self.protocol = protocol
        """Stream communication protocol."""

    async def _send(self, msg: bytes) -> None:
        if self._writer is None:
            raise EOFError("Not connected")
        try:
            await self.protocol.asyncio_send(self._writer, msg)
        except ConnectionError as exc:
            raise EOFError from exc

    async def _receive(self) -> bytes:
        if self._reader is None:
            raise EOFError("Not connected")
        try:
            return await self.protocol.asyncio_receive(self._reader)
        except EOFError:
            if self._writer is not None:
                self._writer.close()
            raise

    async def reset(self) -> None:
        """Reset or establish the connection.

        This method not only resets the existing connection but primarilly
        estableshes the new one.
        """
        if not self.connected:
            self._reader, self._writer = await self._open_connection()
            logger.debug("%s: Connected", self)
        else:
            await super().reset()

    @abc.abstractmethod
    async def _open_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        pass

    @property
    def connected(self) -> bool:
        """Check if client is still connected."""
        return self._writer is not None and not self._writer.is_closing()

    def _disconnect(self) -> None:
        if self._writer is not None:
            self._writer.close()

    async def wait_disconnect(self) -> None:
        """Wait for the client's disconnection."""
        if self._writer is not None:
            with contextlib.suppress(ConnectionError):
                await self._writer.wait_closed()


class RpcServerStream(RpcServer):
    """RPC server listenting for SHV connections for streams."""

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], None | collections.abc.Awaitable[None]
        ],
        protocol: type[RpcTransportProtocol] = RpcProtocolStream,
    ) -> None:
        self.client_connected_cb = client_connected_cb
        """Callbact that is called when new client is connected."""
        self.protocol = protocol
        """Stream communication protocol."""
        self._server: asyncio.Server | None = None
        self._clients: weakref.WeakSet[RpcServerStream.Client] = weakref.WeakSet()

    @abc.abstractmethod
    async def _create_server(self) -> asyncio.Server:
        """Create the server instance."""

    def is_serving(self) -> bool:
        """Check if server is accepting new SHV connections."""
        return self._server is not None and self._server.is_serving()

    async def listen(self) -> None:
        """Start accepting new SHV connections."""
        if self._server is None:
            self._server = await self._create_server()
        await self._server.start_serving()

    async def listen_forewer(self) -> None:
        """Listen and block the calling coroutine until cancelation."""
        if self._server is None:
            self._server = await self._create_server()
        await self._server.serve_forever()

    async def _client_connect(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        client = self.Client(reader, writer, self)
        self._clients.add(client)
        res = self.client_connected_cb(client)
        if isinstance(res, collections.abc.Awaitable):
            await res

    def close(self) -> None:  # noqa D102
        if self._server is not None:
            self._server.close()

    async def wait_closed(self) -> None:  # noqa D102
        if self._server is not None:
            await self._server.wait_closed()

    def terminate(self) -> None:  # noqa D102
        self.close()
        for client in self._clients:
            client.disconnect()

    async def wait_terminated(self) -> None:  # noqa D102
        await self.wait_closed()
        res = await asyncio.gather(
            *(c.wait_disconnect() for c in self._clients),
            return_exceptions=True,
        )
        excs = [v for v in res if isinstance(v, BaseException)]
        if excs:
            if len(excs) == 1:
                raise excs[0]
            raise BaseExceptionGroup("", excs)

    class Client(RpcClient):
        """RPC client for Asyncio's stream server connection."""

        def __init__(
            self,
            reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter,
            server: RpcServerStream,
        ) -> None:
            super().__init__()
            self._reader = reader
            self._writer = writer
            self.protocol = server.protocol
            """Stream communication protocol."""

        async def _send(self, msg: bytes) -> None:
            try:
                await self.protocol.asyncio_send(self._writer, msg)
            except ConnectionError as exc:
                raise EOFError from exc

        async def _receive(self) -> bytes:
            try:
                return await self.protocol.asyncio_receive(self._reader)
            except EOFError:
                self._writer.close()
                raise

        @property
        def connected(self) -> bool:
            """Check if client is still connected."""
            return not self._writer.is_closing()

        def _disconnect(self) -> None:
            self._writer.close()

        async def wait_disconnect(self) -> None:
            """Wait for the client's disconnection."""
            with contextlib.suppress(ConnectionError):
                await self._writer.wait_closed()



"""Check our implementation of fnmatch."""

import pytest

from shv import (
    rpcri_legacy_subscription,
    rpcri_match,
    rpcri_relative_to,
)


@pytest.mark.parametrize(
    "ri,path,method,signal,result",
    (
        (":ls", "", "ls", None, True),
        (":ls", "some", "ls", None, False),
        (":ls", "", "ls", "lsmod", False),
        (":ls:lsmod", "", "ls", "lsmod", True),
        (":*:*", "", "ls", "lsmod", True),
        (":*:*", "", "ls", None, False),
        (":*:lsmod", "", "ls", "lsmod", True),
        (":ls:*", "", "ls", "lsmod", True),
        ("*:ls", "foo", "ls", None, True),
        ("*/*/foo:ls", "one/two/foo", "ls", None, True),
        ("**:ls", "foo", "ls", None, True),
        ("**:ls", "foo/bar/boo", "ls", None, True),
        ("**/boo:ls", "foo/bar/boo", "ls", None, True),
        ("**/bar/**:ls", "foo/bar/boo", "ls", None, True),
        ("**/**/**:ls", "foo/bar/boo", "ls", None, True),
        ("*:ls:lsmod", "foo/faa", "ls", "lsmod", False),
        ("*/*/foo:ls", "one/two/three", "ls", None, False),
        ("**/**:ls", "foo", "ls", None, False),
        ("**/bar:ls", "foo/bar/boo", "ls", None, False),
        ("**/boo/**:ls", "foo/bar/boo", "ls", None, False),
        ("", "", "ls", None, False),
    ),
)
def test_rpcri_match(ri, path, method, signal, result):
    assert rpcri_match(ri, path, method, signal) is result


# @pytest.mark.parametrize(
#    "pattern,path,result",
#    (
#        ("", "", None),
#        ("test/some", "test", "some"),
#        ("test/some", "test/some", None),
#        ("*/*", "foo", "*"),
#        ("*/*", "foo/faa", None),
#        ("**", "foo", "**"),
#        ("foo/**", "foo", "**"),
#        ("some/**", "foo", None),
#        ("foo/**", "foo/bar", "**"),
#        ("foo/**/some", "foo/bar", "**/some"),
#        ("**/some", "foo/bar", "**/some"),
#    ),
# )
# def test_tail_pattern(pattern, path, result):
#    assert shvpath_tail(path, pattern) == result


@pytest.mark.parametrize(
    "ri,path,res",
    (
        ("**:*:*", "", "**:*:*"),
        # ("test/some:*:*chng", "", "test/some:*:*chng"),
        ("test/some:*", "test", "some:*"),
        # ("test/some:*", "test/", "some:*"),
        ("test/some:*", "test/some", None),
        ("test/some:*", "test/some/", None),
        ("test/some:*", "test/som", None),
        ("test/some/*:*", "test/some", "*:*"),
        ("test/some/*:*", "test", "some/*:*"),
        ("test/some/*:*", "tes", None),
        ("test/some/*:*", "test/some/node", None),
        ("**/some/*:*", "test/it/some", "*:*"),
        ("**/some/*:*", "test/it", "**/some/*:*"),
    ),
)
def test_relative_to(ri, path, res):
    assert rpcri_relative_to(ri, path) == res


@pytest.mark.parametrize(
    "ri,shv",
    (
        ("**:*:*", {"method": "", "path": ""}),
        ("test/this:*:*", {"method": "", "paths": "test/this"}),
        ("**:get:*chng", {"path": "", "source": "get", "methods": "*chng"}),
        ("test/this/**:*:*", {"path": "test/this", "method": ""}),
        ("test/this/*:*:*", {"paths": "test/this/*", "method": ""}),
        ("**:*:*chng", {"path": "", "methods": "*chng"}),
    ),
)
def test_to_legacy_subscription(ri, shv):
    assert rpcri_legacy_subscription(ri) == shv


def test_to_legacy_subscription_meth():
    with pytest.raises(ValueError):
        rpcri_legacy_subscription("test/**:*")



"""Check our meta value assignment."""

import datetime
import decimal

import pytest

from shv import (
    SHVBool,
    SHVBytes,
    SHVDatetime,
    SHVDecimal,
    SHVFloat,
    SHVIMap,
    SHVInt,
    SHVList,
    SHVMap,
    SHVMeta,
    SHVNull,
    SHVStr,
    SHVUInt,
    is_shvbool,
    is_shvnull,
    is_shvtype,
    shvmeta_eq,
)

CLASSES: list = [
    (SHVNull, []),
    (SHVBool, [True]),
    (SHVInt, []),
    (SHVUInt, []),
    (SHVFloat, []),
    (SHVBytes, []),
    (SHVStr, []),
    (SHVDatetime, [2020, 1, 2, 3]),
    (SHVDecimal, []),
    (SHVList, []),
    (SHVMap, []),
    (SHVIMap, []),
]


@pytest.mark.parametrize("cls,args", CLASSES)
def test_meta(cls, args):
    """Test our SHVMeta based classes."""
    obj = cls(*args)
    assert obj.meta == {}
    meta = {"foo": 42}
    obj.meta.update(meta)
    assert obj.meta == meta


def test_null():
    """Check that we simulate None somewhat closely."""
    null = SHVNull()
    assert bool(null) == bool(None)
    assert hash(null) == hash(None)


def test_bool():
    """Check that we can compare booleans."""
    false = False
    true = True
    assert SHVBool(True)
    assert not SHVBool(False)
    assert SHVBool(True) == true
    assert SHVBool(True) != false
    assert SHVBool(False) == false
    assert SHVBool(False) != true
    assert SHVBool(True) == SHVBool(True)
    assert SHVBool(False) == SHVBool(False)
    assert SHVBool(False) != SHVBool(True)
    assert SHVBool(True) != SHVBool(False)


REPRS: tuple = (
    (SHVBool, True, False),
    (SHVInt, 1, 42),
    (SHVFloat, 1.0, 4.2),
    (SHVBytes, b"foo", b""),
    (SHVStr, "foo", ""),
    (SHVDecimal, decimal.Decimal("4.2e12"), decimal.Decimal()),
    (SHVList, [1, "foo"], []),
    (SHVMap, {"foo": 1, "": 42}, {}),
    (SHVIMap, {1: "foo", 42: ""}, {}),
)


@pytest.mark.parametrize("cls,value1,value2", REPRS)
def test_repr(cls, value1, value2):
    """Check representation interchange."""
    obj1 = cls(value1)
    obj2 = cls(value2)
    assert obj1 == value1
    assert obj2 == value2
    obj1 = value2
    assert obj1 == obj2


@pytest.mark.parametrize("cls,value1,value2", REPRS)
def test_noteq(cls, value1, value2):
    """Check representation interchange in not equal."""
    obj1 = cls(value1)
    obj2 = cls(value2)
    assert obj2 != value1
    assert obj1 != value2
    assert obj1 != obj2


@pytest.mark.parametrize("cls,value1,value2", REPRS)
def test_hash(cls, value1, value2):
    """Our objects should not modify hash and thus it has to be the same."""
    if cls in {SHVList, SHVMap, SHVIMap}:
        return  # Unhashable types
    assert hash(cls(value1)) == hash(value1)
    assert hash(cls(value2)) == hash(value2)


def test_datetime():
    timestamp = 45322
    shv = SHVDatetime.fromtimestamp(45322)
    base = datetime.datetime.fromtimestamp(timestamp)
    assert shv == base
    assert shv.year == 1970


@pytest.mark.parametrize(
    "value,expected",
    (
        (None, SHVNull()),
        (True, SHVBool(True)),
        (False, SHVBool(False)),
        (42, SHVInt(42)),
        (4.2, SHVFloat(4.2)),
        (b"foo", SHVBytes(b"foo")),
        ("foo", SHVStr("foo")),
        (datetime.datetime.fromtimestamp(352434), SHVDatetime.fromtimestamp(352434)),
        (decimal.Decimal(42), SHVDecimal(42)),
        ([1, "foo"], SHVList([1, "foo"])),
        ({"foo": 1}, SHVMap({"foo": 1})),
        ({1: "foo"}, SHVIMap({1: "foo"})),
    ),
)
def test_new(value, expected):
    obj = SHVMeta.new(value, {})
    assert obj == value
    assert value == obj
    assert obj == expected


@pytest.mark.parametrize(
    "value,expected",
    (
        (None, True),
        (SHVNull(), True),
        (0, False),
    ),
)
def test_is_shvnull(value, expected):
    assert is_shvnull(value) is expected


@pytest.mark.parametrize(
    "value,expected",
    (
        (True, True),
        (False, True),
        (SHVBool(True), True),
        (SHVBool(False), True),
        (0, False),
    ),
)
def test_is_shvbool(value, expected):
    assert is_shvbool(value) is expected


@pytest.mark.parametrize(
    "value,expected",
    (
        (None, True),
        (SHVNull(), True),
        (True, True),
        (False, True),
        (SHVBool(True), True),
        (SHVBool(False), True),
        (42, True),
        (0.42, True),
        (decimal.Decimal("0.42"), True),
        (b"foo", True),
        ("foo", True),
        (datetime.datetime.now(), True),
        ([4, 2], True),
        (["foo", datetime.datetime.now(), None, False], True),
        ({10: 4, 1: 2}, True),
        ({"tenth": 4, "cent": 2}, True),
        ({10: 4, "cent": 2}, False),
        ([{10: 4, 1: [2, 0]}, {"tenth": 4, "cent": 2}], True),
        (datetime.UTC, False),
    ),
)
def test_is_shvtype(value, expected):
    assert is_shvtype(value) is expected


@pytest.mark.parametrize("cls,args", CLASSES)
def test_meta_eq(cls, args):
    """meta_eq should compare metas to each other."""
    obj1 = cls(*args)
    obj2 = cls(*args)
    assert shvmeta_eq(obj1, obj2)
    meta = {"foo": 42}
    obj1.meta.update(meta)
    assert not shvmeta_eq(obj1, obj2)
    obj2.meta.update(meta)
    assert shvmeta_eq(obj1, obj2)


@pytest.mark.parametrize(
    "obj1,obj2",
    (
        (SHVInt(42), 42),
        (SHVNull(), None),
        (SHVNull(), SHVNull()),
    ),
)
def test_meta_eq_eq(obj1, obj2):
    assert shvmeta_eq(obj1, obj2)


@pytest.mark.parametrize("obj1,obj2", ((SHVInt(42), SHVUInt(42)),))
def test_meta_eq_ne(obj1, obj2):
    assert not shvmeta_eq(obj1, obj2)



"""RPC client that represents device in a physical sense."""

import collections.abc

from .rpcmethod import RpcMethodAccess, RpcMethodDesc
from .simpleclient import SimpleClient
from .value import SHVType


class SimpleDevice(SimpleClient):
    """SHV client that represents a physical device."""

    APP_NAME: str = "pyshv-device"

    DEVICE_NAME: str = "unknown"
    """String identifier for the physical device."""

    DEVICE_VERSION: str = "unknown"
    """The physical device design version."""

    DEVICE_SERIAL_NUMBER: str | None = None
    """Serial number of the physical device."""

    async def _method_call(
        self,
        path: str,
        method: str,
        param: SHVType,
        access: RpcMethodAccess,
        user_id: str | None,
    ) -> SHVType:
        match path, method:
            case ".device", "name":
                return self.DEVICE_NAME
            case ".device", "version":
                return self.DEVICE_VERSION
            case ".device", "serialNumber" if access >= RpcMethodAccess.READ:
                return self.DEVICE_SERIAL_NUMBER
        return await super()._method_call(path, method, param, access, user_id)

    def _ls(self, path: str) -> collections.abc.Iterator[str]:
        yield from super()._ls(path)
        if not path:
            yield ".device"

    def _dir(self, path: str) -> collections.abc.Iterator[RpcMethodDesc]:
        yield from super()._dir(path)
        if path == ".device":
            yield RpcMethodDesc.getter(
                "name", "Null", "String", access=RpcMethodAccess.BROWSE
            )
            yield RpcMethodDesc.getter(
                "version", "Null", "String", access=RpcMethodAccess.BROWSE
            )
            yield RpcMethodDesc.getter(
                "serialNumber", "Null", "OptionalString", access=RpcMethodAccess.BROWSE
            )

    async def _lsmod(
        self, path: str, nodes: collections.abc.Mapping[str, bool]
    ) -> None:
        """Report change in the ls method.

        This provides implementation for "lsmod" signal that must be used when
        you are changing the nodes tree to signal clients about that. The
        argument specifies top level nodes added or removed (based on the
        mapping value).

        :param path: SHV path to the valid node which children were added or
          removed.
        :param nodes: Map where key is node name of the node that is top level
          node, that was either added (for value ``True``) or removed (for value
          ``False``).
        """
        await self._signal(path, "lsmod", "ls", nodes, RpcMethodAccess.BROWSE)



"""Connection over serial device."""

import asyncio
import collections.abc
import contextlib
import logging
import pathlib

import aioserial

from .abc import RpcClient, RpcServer
from .stream import (
    RpcProtocolSerialCRC,
    RpcTransportProtocol,
)

try:
    import asyncinotify
except (ImportError, TypeError):  # pragma: no cover
    asyncinotify = None  # type: ignore

logger = logging.getLogger(__name__)


class RpcClientTTY(RpcClient):
    """RPC connection to some SHV peer over serial communication device."""

    def __init__(
        self,
        port: str,
        baudrate: int = 115200,
        protocol: type[RpcTransportProtocol] = RpcProtocolSerialCRC,
    ) -> None:
        super().__init__()
        self.port = port
        self.baudrate = baudrate
        self.protocol = protocol
        self.serial: None | aioserial.AioSerial = None
        self._eof = asyncio.Event()
        self._eof.set()

    def __str__(self) -> str:
        return f"tty:{self.port}"

    async def _send(self, msg: bytes) -> None:
        await self.protocol.send(self._write_async, msg)

    async def _receive(self) -> bytes:
        return await self.protocol.receive(self._read_exactly)

    @property
    def connected(self) -> bool:  # noqa: D102
        return self.serial is not None and self.serial.is_open

    async def reset(self) -> None:  # noqa: D102
        if not self.connected:
            self.serial = aioserial.AioSerial(
                port=self.port,
                baudrate=self.baudrate,
                rtscts=True,
                dsrdtr=True,
                exclusive=True,
            )
            self._eof.clear()
            logger.debug("%s: Connected", self)
        await super().reset()

    async def _write_async(self, data: bytes) -> None:
        assert self.serial is not None
        await self.serial.write_async(data)

    async def _read_exactly(self, n: int) -> bytes:
        assert self.serial is not None
        res = bytearray()
        while len(res) < n:
            try:
                res += await self.serial.read_async(n - len(res))
            except aioserial.SerialException as exc:
                self._eof.set()
                raise EOFError from exc
        return res

    def _disconnect(self) -> None:
        if self.connected:
            assert self.serial is not None
            self.serial.close()
            self._eof.set()

    async def wait_disconnect(self) -> None:  # noqa: D102
        await self._eof.wait()


class RpcServerTTY(RpcServer):
    """RPC server waiting for TTY to appear.

    This actually only maintains a single client as there can't be more than one
    client on single TTY.
    """

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], None | collections.abc.Awaitable[None]
        ],
        port: str,
        baudrate: int = 115200,
        protocol: type[RpcTransportProtocol] = RpcProtocolSerialCRC,
    ) -> None:
        self.client_connected_cb = client_connected_cb
        """Callbact that is called when new client is connected."""
        self.client = RpcClientTTY(port, baudrate, protocol)
        """The :class:`RpcClientTTY` instance."""
        self._task: asyncio.Task | None = None

    def __str__(self) -> str:
        return f"server.tty:{self.client.port}"

    async def _loop(self) -> None:
        while True:
            try:
                await self.client.reset()
            except OSError as exc:
                logger.debug("%s: Waiting for accessible TTY device: %s", self, exc)
            else:
                logger.debug("%s: Openned", self)
                res = self.client_connected_cb(self.client)
                if isinstance(res, collections.abc.Awaitable):
                    await res
                await self.client.wait_disconnect()
                continue
            if asyncinotify is not None:
                with asyncinotify.Inotify() as inotify:
                    pth = pathlib.Path(self.client.port)
                    inotify.add_watch(
                        pth.parent, asyncinotify.Mask.CREATE | asyncinotify.Mask.ATTRIB
                    )
                    async for event in inotify:
                        if str(pth.name) == str(event.name):
                            break
            else:
                await asyncio.sleep(5)  # type: ignore

    def is_serving(self) -> bool:  # noqa: D102
        return self._task is not None and not self._task.done()

    async def listen(self) -> None:  # noqa: D102
        if not self.is_serving():
            self._task = asyncio.create_task(self._loop())

    async def listen_forewer(self) -> None:  # noqa: D102
        await self.listen()
        await self.wait_closed()

    def close(self) -> None:  # noqa: D102
        if self._task is not None:
            self._task.cancel()

    async def wait_closed(self) -> None:  # noqa: D102
        if self._task is not None:
            with contextlib.suppress(asyncio.CancelledError):
                await self._task

    def terminate(self) -> None:  # noqa D102
        self.close()
        self.client.disconnect()

    async def wait_terminated(self) -> None:  # noqa D102
        await self.wait_closed()
        await self.client.wait_disconnect()



"""Check usability of our stream conversion."""

import pytest

from shv.cpconv import CPFormat, convert

DATA = [
    ("true", b"\xfe"),
    ("[1]", b"\x88A\xff"),
]


@pytest.mark.parametrize("cpon,chainpack", DATA)
def test_cpconv(cpon, chainpack):
    assert (
        convert(chainpack, CPFormat.CHAINPACK, None, CPFormat.CPON).decode("utf-8")
        == cpon
    )


@pytest.mark.parametrize("cpon,chainpack", DATA)
def test_cpon2cp(cpon, chainpack):
    assert convert(cpon, CPFormat.CPON, None, CPFormat.CHAINPACK) == chainpack



"""Connection over Pipes or socket."""

from __future__ import annotations

import asyncio
import logging
import os
import typing

from .stream import (
    RpcClientStream,
    RpcProtocolSerial,
    RpcTransportProtocol,
)

logger = logging.getLogger(__name__)


class RpcClientPipe(RpcClientStream):
    """RPC connection to some SHV peer over Unix pipes or other such streams."""

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        protocol: type[RpcTransportProtocol] = RpcProtocolSerial,
    ) -> None:
        super().__init__(protocol)
        self._reader = reader
        self._writer = writer

    def __str__(self) -> str:
        return "pipe"

    async def _open_connection(  # noqa PLR6301
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        raise ConnectionError("Pipes can't be reconnected")

    @classmethod
    async def fdopen(
        cls,
        rpipe: int | typing.IO,
        wpipe: int | typing.IO,
        protocol: type[RpcTransportProtocol] = RpcProtocolSerial,
    ) -> RpcClientPipe:
        """Create RPC client from existing Unix pipes."""
        if isinstance(rpipe, int):
            rpipe = os.fdopen(rpipe, mode="r")
        reader = asyncio.StreamReader()
        srprotocol = asyncio.StreamReaderProtocol(reader)
        await asyncio.get_running_loop().connect_read_pipe(lambda: srprotocol, rpipe)

        if isinstance(wpipe, int):
            wpipe = os.fdopen(wpipe, mode="w")
        wtransport, _ = await asyncio.get_running_loop().connect_write_pipe(
            lambda: srprotocol, wpipe
        )
        writer = asyncio.StreamWriter(
            wtransport, srprotocol, None, asyncio.get_running_loop()
        )

        return cls(reader, writer, protocol)

    @classmethod
    async def open_pair(
        cls,
        protocol: type[RpcTransportProtocol] = RpcProtocolSerial,
        flags: int = 0,
    ) -> tuple[RpcClientPipe, RpcClientPipe]:
        """Create pair of clients that are interconnected over the pipe.

        :param protocol: The protocol factory to be used.
        :param flags: Flags passed to :meth:`os.pipe2`.
        :return: Pair of clients that are interconnected over Unix pipes.
        """
        pr1, pw1 = os.pipe2(flags)
        pr2, pw2 = os.pipe2(flags)
        client1 = await cls.fdopen(pr1, pw2, protocol)
        client2 = await cls.fdopen(pr2, pw1, protocol)
        return client1, client2



"""Common base for SHV writers and readers."""

from __future__ import annotations

import abc
import asyncio
import collections.abc
import datetime
import decimal
import io
import typing

from .value import (
    SHVIMapType,
    SHVListType,
    SHVMapType,
    SHVMetaType,
    SHVType,
    SHVUInt,
    is_shvbool,
    is_shvimap,
    is_shvmap,
    is_shvnull,
    shvmeta,
)


class CommonReader(abc.ABC):
    """Common reader base."""

    def __init__(self, stream: bytes | bytearray | typing.IO | CommonReader) -> None:
        if isinstance(stream, bytes | bytearray):
            stream = io.BytesIO(stream)
        if isinstance(stream, CommonReader):
            orig = stream
            stream = stream.stream

        self.stream: typing.IO = stream
        """Stream used to receive bytes from."""
        self.peek_bytes = b""
        """Peeked bytes so far."""
        self.bytes_cnt = 0
        """Number of bytes read so far."""

        if isinstance(stream, CommonReader):
            self.peek_bytes = orig.peek_bytes
            self.bytes_cnt = orig.bytes_cnt

    def _read(self, size: int) -> bytes:
        """Read bytes from the stream.

        Compared to regular read this reads always given number of bytes.

        :param size: Number of bytes to read.
        :return: Read bytes.
        :raise EOFError: in case EOF is encountered.
        """
        assert size > 0
        res = b""
        while len(res) < size:
            read = self.read_raw(size - len(res))
            if len(read) == 0:
                raise EOFError("End of message or file encountered.")
            res += read
        self.bytes_cnt += size
        return res

    def _read_byte(self) -> int:
        """Read a single byte and return it as int.

        :return: Read byte.
        :raise EOFError: in case EOF is encountered.
        """
        return self._read(1)[0]

    def _peek_byte(self) -> int:
        """Peek a single byte and return it as int.

        Peeked byte is actually read from the stream and is stored to
        `peek_bytes`.

        :return: Peeked byte. This also returns `0` when EOF is encountered.
        Make sure that you use `0` to terminate what you are doing.
        """
        if len(self.peek_bytes) == 0:
            try:
                self.peek_bytes = self._read(1)
                self.bytes_cnt -= len(self.peek_bytes)
            except EOFError:
                return 0
        return self.peek_bytes[0]

    def _peek_drop(self) -> None:
        """Drop peeked data."""
        self.bytes_cnt += len(self.peek_bytes)
        self.peek_bytes = b""

    def _read_check(self, data: bytes) -> None:
        """Read bytes and check that it is what we expected.

        :param data: Expected bytes.
        :raise ValueError: when unexpected byte was received.
        :raise EOFError: in case EOF is encountered.
        """
        b = self._read(len(data))
        if data != b:
            raise ValueError(f"Expected {data!r} but got {b!r}")

    def read_raw(self, size: int) -> bytes:
        """Read raw bytes from the stream."""
        res = self.peek_bytes[:size]
        self.peek_bytes = self.peek_bytes[size:]
        if len(res) < size:
            res += self.stream.read(size - len(res))
        return res

    @abc.abstractmethod
    def read(self) -> SHVType:
        """Read next SHV value.

        :raise ValueError: when unexpected byte was received.
        :raise EOFError: in case EOF is encountered.
        """

    @abc.abstractmethod
    def read_meta(self) -> SHVMetaType | None:
        """Read next meta value without reading the value.

        :raise EOFError: in case EOF is encountered.
        """

    @classmethod
    def unpack(cls, data: bytes | str) -> SHVType:
        """Unpack single value from given data."""
        if isinstance(data, str):
            data = data.encode("utf-8")
        self = cls(io.BytesIO(data))
        return self.read()


class CommonWriter(abc.ABC):
    """Common writer base."""

    def __init__(self, stream: typing.IO | asyncio.StreamWriter | None = None) -> None:
        self.stream = stream or io.BytesIO()
        self.bytes_cnt = 0

    def _write(self, data: bytes | int) -> None:
        if isinstance(data, int):
            data = bytes([data])
        self.bytes_cnt += len(data)
        self.stream.write(data)

    def _writestr(self, data: str) -> None:
        self._write(data.encode("utf-8"))

    def write(self, value: SHVType) -> None:
        """Write generic RpcValue."""
        if shvmeta(value):
            self.write_meta(shvmeta(value))
        if is_shvnull(value):
            self.write_null()
        elif is_shvbool(value):
            self.write_bool(bool(value))
        elif isinstance(value, SHVUInt):
            self.write_uint(value)
        elif isinstance(value, int):
            self.write_int(value)
        elif isinstance(value, float):
            self.write_double(value)
        elif isinstance(value, decimal.Decimal):
            self.write_decimal(value)
        elif isinstance(value, bytes):
            self.write_blob(value)
        elif isinstance(value, str):
            self.write_string(value)
        elif isinstance(value, datetime.datetime):
            self.write_datetime(value)
        elif isinstance(value, collections.abc.Sequence):
            self.write_list(value)
        elif is_shvimap(value):
            self.write_imap(value)
        elif is_shvmap(value):
            self.write_map(value)
        else:
            raise ValueError(f"Invalid value for SHV: {value!r}")

    @abc.abstractmethod
    def write_meta(self, meta: SHVMetaType) -> None:
        """Write given meta to the stream."""

    @abc.abstractmethod
    def write_null(self) -> None:
        """Write null to the stream."""

    @abc.abstractmethod
    def write_bool(self, value: bool) -> None:
        """Write boolean to the stream."""

    @abc.abstractmethod
    def write_blob(self, value: bytes | bytearray) -> None:
        """Write blob to the stream."""

    @abc.abstractmethod
    def write_string(self, value: str) -> None:
        """Write string to the stream."""

    @abc.abstractmethod
    def write_cstring(self, value: str) -> None:
        """Write C string to the stream."""

    @abc.abstractmethod
    def write_uint(self, value: int) -> None:
        """Write unsigned integer to the stream."""

    @abc.abstractmethod
    def write_int(self, value: int) -> None:
        """Write signed integer to the stream."""

    @abc.abstractmethod
    def write_double(self, value: float) -> None:
        """Write floating point number to the stream."""

    @abc.abstractmethod
    def write_decimal(self, value: decimal.Decimal) -> None:
        """Write decimal number to the stream."""

    @abc.abstractmethod
    def write_list(self, value: SHVListType) -> None:
        """Write list of other types to the stream."""

    @abc.abstractmethod
    def write_map(self, value: SHVMapType) -> None:
        """Write map of other types to the stream."""

    @abc.abstractmethod
    def write_imap(self, value: SHVIMapType) -> None:
        """Write integer map of other types to the stream."""

    @abc.abstractmethod
    def write_datetime(self, value: datetime.datetime) -> None:
        """Write date and time to the stream."""

    @classmethod
    def pack(cls, value: SHVType) -> bytes:
        """Pack given value and return bytes."""
        stream = io.BytesIO()
        self = cls(stream)
        self.write(value)
        return stream.getvalue()



"""Check that errors can be created trough RpcError."""

import pytest

from shv import (
    RpcError,
    RpcInternalError,
    RpcInvalidParamError,
    RpcInvalidRequestError,
    RpcMethodCallCancelledError,
    RpcMethodCallExceptionError,
    RpcMethodCallTimeoutError,
    RpcMethodNotFoundError,
    RpcParseError,
)

errors = (
    RpcInternalError,
    RpcInvalidParamError,
    RpcInvalidRequestError,
    RpcMethodCallCancelledError,
    RpcMethodCallExceptionError,
    RpcMethodCallTimeoutError,
    RpcMethodNotFoundError,
    RpcParseError,
)


@pytest.mark.parametrize("cls", (RpcError, *errors))
def test_new_error(cls):
    obj = RpcError("foo", cls.shv_error_code)
    assert isinstance(obj, cls)
    assert obj.error_code == cls.shv_error_code
    assert obj.message == "foo"


@pytest.mark.parametrize("cls", errors)
def test_new_direct_error(cls):
    obj = cls("foo")
    assert isinstance(obj, RpcError)
    assert obj.error_code == cls.shv_error_code
    assert obj.message == "foo"



"""Check that we correctly parse and serialize our URL format."""

import re

import pytest

from shv import RpcLogin, RpcLoginType, RpcProtocol, RpcUrl

DATA = [
    ("unix:/dev/null", RpcUrl("/dev/null", protocol=RpcProtocol.UNIX)),
    (
        "unixs:/dev/null?user=test&password=foo",
        RpcUrl(
            "/dev/null",
            protocol=RpcProtocol.UNIXS,
            login=RpcLogin(username="test", password="foo"),
        ),
    ),
    (
        "serial:/dev/null?user=test%40example.com&password=a%C4%8D%C5%A1f",
        RpcUrl(
            "/dev/null",
            protocol=RpcProtocol.TTY,
            login=RpcLogin(username="test@example.com", password="ačšf"),
        ),
    ),
    (
        "tcp://test@localhost:4242",
        RpcUrl("localhost", port=4242, login=RpcLogin(username="test")),
    ),
    (
        "ssl://test@localhost:4242",
        RpcUrl(
            "localhost",
            protocol=RpcProtocol.SSL,
            port=4242,
            login=RpcLogin(username="test"),
        ),
    ),
    (
        "tcp://localhost:4242?devid=foo&devmount=/dev/null",
        RpcUrl(
            "localhost",
            port=4242,
            login=RpcLogin(
                options={"device": {"deviceId": "foo", "mountPoint": "/dev/null"}}
            ),
        ),
    ),
    (
        "tcp://localhost:4242?devid=foo&devmount=/dev/null&password=test",
        RpcUrl(
            "localhost",
            port=4242,
            login=RpcLogin(
                password="test",
                login_type=RpcLoginType.PLAIN,
                options={"device": {"deviceId": "foo", "mountPoint": "/dev/null"}},
            ),
        ),
    ),
    (
        "tcp://localhost:4242?devid=foo&devmount=/dev/null&shapass=" + ("x" * 40),
        RpcUrl(
            "localhost",
            port=4242,
            login=RpcLogin(
                password="x" * 40,
                login_type=RpcLoginType.SHA1,
                options={"device": {"deviceId": "foo", "mountPoint": "/dev/null"}},
            ),
        ),
    ),
    ("tcp://[::]:4242", RpcUrl("::", port=4242)),
    ("serial:/dev/ttyX", RpcUrl("/dev/ttyX", protocol=RpcProtocol.TTY)),
    (
        "serial:/dev/ttyX?baudrate=1152000",
        RpcUrl("/dev/ttyX", protocol=RpcProtocol.TTY, baudrate=1152000),
    ),
]


@pytest.mark.parametrize("url,rpcurl", DATA)
def test_to_url(url, rpcurl):
    """Check that we correctly parse these URLs."""
    assert rpcurl.to_url() == url


@pytest.mark.parametrize(
    "url,rpcurl",
    [
        *DATA,
        ("", RpcUrl("", protocol=RpcProtocol.UNIX)),
        ("/dev/null", RpcUrl("/dev/null", protocol=RpcProtocol.UNIX)),
        ("//dev/null", RpcUrl("/dev/null", protocol=RpcProtocol.UNIX)),
        (
            "unix://dev/null",
            RpcUrl("/dev/null", protocol=RpcProtocol.UNIX),
        ),
        ("tcp://localhost", RpcUrl("localhost", port=3755, protocol=RpcProtocol.TCP)),
        ("tcps://localhost", RpcUrl("localhost", port=3765, protocol=RpcProtocol.TCPS)),
        ("ssl://localhost", RpcUrl("localhost", port=3756, protocol=RpcProtocol.SSL)),
        ("ssls://localhost", RpcUrl("localhost", port=3766, protocol=RpcProtocol.SSLS)),
        (
            "tcp://test@localhost:4242",
            RpcUrl("localhost", port=4242, login=RpcLogin(username="test")),
        ),
        (
            "tcp://localhost?devid=foo",
            RpcUrl(
                "localhost", login=RpcLogin(options={"device": {"deviceId": "foo"}})
            ),
        ),
    ],
)
def test_parse(url, rpcurl):
    """Check that we correctly parse these URLs."""
    assert RpcUrl.parse(url) == rpcurl


@pytest.mark.parametrize(
    "url,msg",
    (
        ("tcp://localhost?missing=some", "Unsupported URL queries: missing"),
        (
            "tcp://localhost?missing=some&other=foo",
            "Unsupported URL queries: missing, other",
        ),
        ("tcp:///dev/null", "Path is not supported for tcp: /dev/null"),
    ),
)
def test_invalid(url, msg):
    with pytest.raises(ValueError, match=re.escape(msg)):
        RpcUrl.parse(url)



"""Various tools used in the code."""

import collections.abc
import typing

T = typing.TypeVar("T")


def lookahead(
    iterin: collections.abc.Iterable[T],
) -> collections.abc.Iterator[tuple[T, bool]]:
    """Itearte and tell if there is more data comming."""
    it = iter(iterin)
    try:
        prev = next(it)
    except StopIteration:
        return
    for v in it:
        yield prev, True
        prev = v
    yield prev, False



"""Functions providing helpers for client creation based on the RPC URL."""

import collections.abc

from ..rpcurl import RpcProtocol, RpcUrl
from .abc import RpcClient, RpcServer
from .stream import RpcProtocolSerial, RpcProtocolSerialCRC, RpcProtocolStream
from .tcp import RpcClientTCP, RpcServerTCP
from .tty import RpcClientTTY, RpcServerTTY
from .unix import RpcClientUnix, RpcServerUnix
from .ws import RpcClientWebSockets, RpcServerWebSockets, RpcServerWebSocketsUnix


def init_rpc_client(url: RpcUrl | str) -> RpcClient:
    """Initialize correct :class:`RpcClient` for given URL.

    :param url: RPC URL specifying the connection target.
    :return: Chosen :class:`RpcClient` child instance based on the passed URL.
    """
    if isinstance(url, str):
        url = RpcUrl.parse(url)
    match url.protocol:
        case RpcProtocol.TCP:
            return RpcClientTCP(url.location, url.port, RpcProtocolStream)
        case RpcProtocol.TCPS:
            return RpcClientTCP(url.location, url.port, RpcProtocolSerial)
        case RpcProtocol.UNIX:
            return RpcClientUnix(url.location, RpcProtocolStream)
        case RpcProtocol.UNIXS:
            return RpcClientUnix(url.location, RpcProtocolSerial)
        case RpcProtocol.TTY:
            return RpcClientTTY(url.location, url.baudrate, RpcProtocolSerialCRC)
        case RpcProtocol.WS:
            return RpcClientWebSockets(url.location)
        case _:
            raise NotImplementedError(f"Unimplemented protocol: {url.protocol}")


async def connect_rpc_client(url: RpcUrl | str) -> RpcClient:
    """Initialize and establish :class:`RpcClient` connection for given URL.

    Compared to the :func:`init_rpc_client` this also calls
    :meth:`RpcClient.reset` to establish the connection.

    :param url: RPC URL specifying the connection target.
    :return: Chosen :class:`RpcClient` child instance based on the passed URL.
    """
    res = init_rpc_client(url)
    await res.reset()
    return res


async def create_rpc_server(
    client_connected_cb: collections.abc.Callable[
        [RpcClient], None | collections.abc.Awaitable[None]
    ],
    url: RpcUrl | str,
) -> RpcServer:
    """Create server listening on given URL.

    :param client_connected_cb: function called for every new client connected.
    :param url: RPC URL specifying where server should listen.
    """
    res: RpcServer
    if isinstance(url, str):
        url = RpcUrl.parse(url)
    match url.protocol:
        case RpcProtocol.TCP:
            res = RpcServerTCP(
                client_connected_cb, url.location, url.port, RpcProtocolStream
            )
        case RpcProtocol.TCPS:
            res = RpcServerTCP(
                client_connected_cb, url.location, url.port, RpcProtocolSerial
            )
        case RpcProtocol.UNIX:
            res = RpcServerUnix(client_connected_cb, url.location, RpcProtocolStream)
        case RpcProtocol.UNIXS:
            res = RpcServerUnix(client_connected_cb, url.location, RpcProtocolSerial)
        case RpcProtocol.TTY:
            res = RpcServerTTY(
                client_connected_cb, url.location, url.baudrate, RpcProtocolSerialCRC
            )
        case RpcProtocol.WS:
            if url.port != -1:
                res = RpcServerWebSockets(client_connected_cb, url.location, url.port)
            else:
                res = RpcServerWebSocketsUnix(client_connected_cb, url.location)
        case _:
            raise NotImplementedError(f"Unimplemented protocol: {url.protocol}")
    await res.listen()
    return res



"""Python implementation of Silicon Heaven."""

from .__version__ import VERSION
from .chainpack import ChainPack, ChainPackReader, ChainPackWriter
from .cpon import Cpon, CponReader, CponWriter
from .rpcerrors import (
    RpcError,
    RpcErrorCode,
    RpcInternalError,
    RpcInvalidParamError,
    RpcInvalidRequestError,
    RpcLoginRequiredError,
    RpcMethodCallCancelledError,
    RpcMethodCallExceptionError,
    RpcMethodCallTimeoutError,
    RpcMethodNotFoundError,
    RpcParseError,
    RpcUserIDRequiredError,
)
from .rpclogin import RpcLogin, RpcLoginType
from .rpcmessage import RpcMessage
from .rpcmethod import RpcMethodAccess, RpcMethodDesc, RpcMethodFlags
from .rpcparam import SHVGetKey, shvarg, shvargt, shvget, shvgett, shvt
from .rpcri import rpcri_legacy_subscription, rpcri_match, rpcri_relative_to
from .rpctransport import (
    RpcClient,
    RpcClientPipe,
    RpcClientTCP,
    RpcClientTTY,
    RpcClientUnix,
    RpcClientWebSockets,
    RpcProtocolSerial,
    RpcProtocolSerialCRC,
    RpcProtocolStream,
    RpcServer,
    RpcServerTCP,
    RpcServerTTY,
    RpcServerUnix,
    RpcServerWebSockets,
    RpcServerWebSocketsUnix,
    RpcTransportProtocol,
    connect_rpc_client,
    create_rpc_server,
    init_rpc_client,
)
from .rpcurl import RpcProtocol, RpcUrl
from .shvversion import SHV_VERSION_MAJOR, SHV_VERSION_MINOR
from .simplebase import SimpleBase
from .simpleclient import SimpleClient
from .simpledevice import SimpleDevice
from .value import (
    SHVBool,
    SHVBoolType,
    SHVBytes,
    SHVDatetime,
    SHVDecimal,
    SHVFloat,
    SHVIMap,
    SHVIMapType,
    SHVInt,
    SHVList,
    SHVListType,
    SHVMap,
    SHVMapType,
    SHVMeta,
    SHVMetaType,
    SHVNull,
    SHVNullType,
    SHVStr,
    SHVType,
    SHVUInt,
    decimal_rexp,
    is_shvbool,
    is_shvimap,
    is_shvmap,
    is_shvnull,
    is_shvtype,
    shvmeta,
    shvmeta_eq,
)
from .valueclient import ValueClient

__all__ = [
    "SHV_VERSION_MAJOR",
    "SHV_VERSION_MINOR",
    "VERSION",
    "ChainPack",
    "ChainPackReader",
    "ChainPackWriter",
    "Cpon",
    "CponReader",
    "CponWriter",
    "RpcClient",
    "RpcClientPipe",
    "RpcClientTCP",
    "RpcClientTTY",
    "RpcClientUnix",
    "RpcClientWebSockets",
    "RpcError",
    "RpcErrorCode",
    "RpcInternalError",
    "RpcInvalidParamError",
    "RpcInvalidRequestError",
    "RpcLogin",
    "RpcLoginRequiredError",
    "RpcLoginType",
    "RpcLoginType",
    "RpcMessage",
    "RpcMethodAccess",
    "RpcMethodCallCancelledError",
    "RpcMethodCallExceptionError",
    "RpcMethodCallTimeoutError",
    "RpcMethodDesc",
    "RpcMethodFlags",
    "RpcMethodNotFoundError",
    "RpcParseError",
    "RpcProtocol",
    "RpcProtocolSerial",
    "RpcProtocolSerialCRC",
    "RpcProtocolStream",
    "RpcServer",
    "RpcServerTCP",
    "RpcServerTTY",
    "RpcServerUnix",
    "RpcServerWebSockets",
    "RpcServerWebSocketsUnix",
    "RpcTransportProtocol",
    "RpcUrl",
    "RpcUserIDRequiredError",
    "SHVBool",
    "SHVBoolType",
    "SHVBytes",
    "SHVDatetime",
    "SHVDecimal",
    "SHVFloat",
    "SHVGetKey",
    "SHVIMap",
    "SHVIMapType",
    "SHVInt",
    "SHVList",
    "SHVListType",
    "SHVMap",
    "SHVMapType",
    "SHVMeta",
    "SHVMetaType",
    "SHVNull",
    "SHVNullType",
    "SHVStr",
    "SHVType",
    "SHVUInt",
    "SimpleBase",
    "SimpleClient",
    "SimpleDevice",
    "ValueClient",
    "connect_rpc_client",
    "create_rpc_server",
    "decimal_rexp",
    "init_rpc_client",
    "is_shvbool",
    "is_shvimap",
    "is_shvmap",
    "is_shvnull",
    "is_shvtype",
    "rpcri_legacy_subscription",
    "rpcri_match",
    "rpcri_relative_to",
    "shvarg",
    "shvargt",
    "shvget",
    "shvgett",
    "shvmeta",
    "shvmeta_eq",
    "shvt",
]



"""Test ValueClient specific tools."""

import asyncio

import pytest


async def test_prop_cache(value_client, example_device):
    """Check that we correctly cache values in value client."""
    await value_client.subscribe("test/device/track/**:*:*")
    assert len(value_client) == 0
    assert not list(iter(value_client))
    assert "test/device/track/1" not in value_client
    assert await value_client.prop_get("test/device/track/1", 8) == [0]
    assert len(value_client) == 1
    assert value_client["test/device/track/1"] == [0]
    assert await value_client.prop_get("test/device/track/1", 8) == [0]
    await value_client.prop_set("test/device/track/1", [1, 2])
    assert value_client["test/device/track/1"] == [1, 2]
    await value_client.unsubscribe("test/device/track/**:*:*")
    assert len(value_client) == 0


async def test_unsubscribe_keep_cache(value_client, example_device):
    """Check that unsubscribe can keep the cached values."""
    await value_client.subscribe("test/device/**:*:*")
    await value_client.prop_get("test/device/track/1")
    assert "test/device/track/1" in value_client
    await value_client.unsubscribe("test/device/**:*:*", clean_cache=False)
    assert "test/device/track/1" in value_client


async def test_is_subscribed(value_client, example_device):
    """Check that we correctly trak our own subscriptions."""
    assert not value_client.is_subscribed("test/foo")
    await value_client.subscribe("test/**:*:*")
    assert value_client.is_subscribed("test/foo")
    assert value_client.is_subscribed("test/device")
    assert value_client.is_subscribed("test/device/any/other/path/under")
    assert not value_client.is_subscribed(".broker")
    assert not value_client.is_subscribed("tester")  # A different node
    await value_client.unsubscribe("test/**:*:*")
    assert not value_client.is_subscribed("test")


async def test_get_snapshot(value_client, example_device):
    """Check ability of get_snapshot to cache required properties."""
    await value_client.subscribe("test/**:*:*")
    await value_client.get_snapshot("test/device")
    assert "test/device/track/1" in value_client


async def test_get_snapshot_subscribes(value_client, example_device):
    """Check get_snapshot without arguments."""
    await value_client.subscribe("test/device/**:*:*")
    await value_client.get_snapshot()
    assert "test/device/track/1" in value_client
    assert len(value_client) == 8


async def test_get_snapshot_lower_subscribe(value_client, example_device):
    """Check that get_snapshot caches only subscribed paths."""
    await value_client.subscribe("test/device/track/2:*:*")
    await value_client.get_snapshot("test")
    assert len(value_client) == 1
    assert "test/device/track/2" in value_client


async def test_wait_for_change(value_client, example_device):
    """Check that simple wait for change notification works."""
    await value_client.subscribe("test/device/**:*:*")
    task1 = asyncio.create_task(value_client.wait_for_change("test/device/track/1"))
    task2 = asyncio.create_task(value_client.wait_for_change("test/device/track/1"))
    await value_client.prop_set("test/device/track/1", [1, 3])
    assert await task1 == [1, 3]
    assert await task2 == [1, 3]


@pytest.mark.parametrize("path", ("test/device", "test", ""))
async def test_prop_change(value_client, example_device, path):
    """Check hook for property change."""
    res = []

    def hook(client, path, value):
        res.append((client, path, value))

    await value_client.subscribe("test/device/**:*:*")
    value_client.on_change(path, hook)
    await value_client.prop_set("test/device/track/1", [1, 2])
    await value_client.prop_set("test/device/track/4", [])
    value_client.on_change(path, None)
    await value_client.prop_set("test/device/track/3", [])

    assert res == [
        (value_client, "test/device/track/1", [1, 2]),
        (value_client, "test/device/track/4", []),
    ]


async def test_prop_change_wait(value_client, example_device):
    """Check that we can wait for property change with prop_change_wait."""
    task1 = asyncio.create_task(value_client.prop_change_wait("test/device/track/1"))
    task2 = asyncio.create_task(value_client.prop_change_wait("test/device/track/1"))
    await value_client.prop_set("test/device/track/1", [1, 3])
    assert await task1 == [1, 3]
    assert await task2 == [1, 3]


async def test_prop_change_subscribed(value_client, example_device):
    """Check that we can wait for property change with prop_change_wait."""
    await value_client.subscribe("test/device/**:*:*")
    task1 = asyncio.create_task(value_client.prop_change_wait("test/device/track/1"))
    task2 = asyncio.create_task(value_client.prop_change_wait("test/device/track/1"))
    await value_client.prop_set("test/device/track/1", [1, 3])
    assert await task1 == [1, 3]
    assert await task2 == [1, 3]


async def test_prop_change_timeout(value_client, example_device):
    """Check that we can will timeout from prop_change_wait."""
    await value_client.subscribe("test/device/**:*:*")
    await value_client.prop_get("test/device/track/1")  # seed cache
    with pytest.raises(TimeoutError):
        print(
            await value_client.prop_change_wait(
                "test/device/track/1", get_period=0.2, timeout=1
            )
        )



"""Chainpack data format reader and writer."""

import datetime
import decimal
import io
import struct
import typing

from . import commonpack
from .value import (
    SHVIMapType,
    SHVListType,
    SHVMapType,
    SHVMeta,
    SHVMetaType,
    SHVType,
    SHVUInt,
    decimal_rexp,
)


class ChainPack:
    """ChainPack constans and definitions."""

    ProtocolType = 1

    CP_Null = 128
    CP_UInt = 129
    CP_Int = 130
    CP_Double = 131
    CP_Bool = 132
    CP_Blob = 133
    CP_String = 134  # utf8 encoded string
    CP_List = 136
    CP_Map = 137
    CP_IMap = 138
    CP_MetaMap = 139
    CP_Decimal = 140
    CP_DateTime = 141
    CP_CString = 142
    CP_FALSE = 253
    CP_TRUE = 254
    CP_TERM = 255

    # UTC msec since 2.2. 2018
    # Fri Feb 02 2018 00:00:00 == 1517529600 EPOCH
    SHV_EPOCH_SEC = 1517529600
    # ChainPack.INVALID_MIN_OFFSET_FROM_UTC = (-64 * 15)

    @staticmethod
    def unpack(data: bytes | str) -> SHVType:
        """Unpack single value from given data."""
        return ChainPackReader.unpack(data)

    @staticmethod
    def pack(value: SHVType) -> bytes:
        """Pack given value and return bytes."""
        return ChainPackWriter.pack(value)

    @staticmethod
    def unpack_uint_data(data: bytes | bytearray) -> int:
        """Unpack given value as unsigned int data.

        The existence of this is pretty much just to support Stream transport
        protocol for SHV RPC. You probably do not want to use this for anything
        else.

        :param data: Data that should contain unsigned integer in Chainpack
          format. Unused bytes at the end are ignored.
        :return: unpacked integer.
        :raise ValueError: in case of invalid data.
        """
        try:
            return ChainPackReader(data).read_uint_data()
        except EOFError as exc:
            raise ValueError from exc

    @staticmethod
    def pack_uint_data(value: int) -> bytes:
        """Pack given value as unsigned int data.

        The existence of this is pretty much just to support Stream transport
        protocol for SHV RPC. You probably do not want to use this for anything
        else.

        :param value: Unsigned integer to be packed.
        :return: bytes with unsigned integer.
        """
        bio = io.BytesIO()
        writer = ChainPackWriter(bio)
        writer.write_uint_data(value)
        return bio.getvalue()


class ChainPackReader(commonpack.CommonReader):
    """Read data in ChainPack format."""

    def read_meta(self) -> SHVMetaType | None:  # noqa: D102
        if self._peek_byte() != ChainPack.CP_MetaMap:
            return None
        self._peek_drop()
        return self._read_map()

    def read(self) -> SHVType:  # noqa: D102
        meta = self.read_meta()

        value: SHVType
        packing_schema = self._read_byte()
        if packing_schema < 128:
            # tiny Int or UInt
            value = packing_schema & 63
            if not packing_schema & 64:
                value = SHVUInt(value)
        elif packing_schema == ChainPack.CP_Null:
            value = None
        elif packing_schema == ChainPack.CP_TRUE:
            value = True
        elif packing_schema == ChainPack.CP_FALSE:
            value = False
        elif packing_schema == ChainPack.CP_Int:
            value = self._read_int_data()
        elif packing_schema == ChainPack.CP_UInt:
            value = SHVUInt(self.read_uint_data())
        elif packing_schema == ChainPack.CP_Double:
            value = self._read_double()
        elif packing_schema == ChainPack.CP_Decimal:
            value = self._read_decimal()
        elif packing_schema == ChainPack.CP_DateTime:
            value = self._read_datetime()
        elif packing_schema == ChainPack.CP_Map:
            value = typing.cast(SHVMapType, self._read_map())
        elif packing_schema == ChainPack.CP_IMap:
            value = typing.cast(SHVIMapType, self._read_map())
        elif packing_schema == ChainPack.CP_List:
            value = self._read_list()
        elif packing_schema == ChainPack.CP_Blob:
            value = self._read_blob()
        elif packing_schema == ChainPack.CP_String:
            value = self._read_string()
        elif packing_schema == ChainPack.CP_CString:
            value = self._read_cstring()
        else:
            raise ValueError(f"ChainPack - Invalid type: {packing_schema}")
        if meta is not None:
            value = SHVMeta.new(value, meta)
        return value

    def _read_uint_data_helper(self) -> tuple[int, int]:
        num = 0
        bitlen = 0
        head = self._read_byte()
        if (head & 128) == 0:
            bytes_to_read_cnt = 0
            num = head & 127
            bitlen = 7
        elif (head & 64) == 0:
            bytes_to_read_cnt = 1
            num = head & 63
            bitlen = 6 + 8
        elif (head & 32) == 0:
            bytes_to_read_cnt = 2
            num = head & 31
            bitlen = 5 + 2 * 8
        elif (head & 16) == 0:
            bytes_to_read_cnt = 3
            num = head & 15
            bitlen = 4 + 3 * 8
        else:
            bytes_to_read_cnt = (head & 0xF) + 4
            bitlen = bytes_to_read_cnt * 8

        for _ in range(bytes_to_read_cnt):
            r = self._read_byte()
            num = (num << 8) + r
        return num, bitlen

    def read_uint_data(self) -> int:  # noqa: D102
        num, _ = self._read_uint_data_helper()
        return num

    def _read_int_data(self) -> int:
        num, bitlen = self._read_uint_data_helper()
        sign_bit_mask = 1 << (bitlen - 1)
        neg = num & sign_bit_mask
        snum = num
        if neg:
            snum &= ~sign_bit_mask
            snum = -snum
        return snum

    def _read_double(self) -> float:
        res = struct.unpack("<d", self._read(8))  # little endian
        return typing.cast(float, res[0])

    def _read_decimal(self) -> decimal.Decimal:
        mant = self._read_int_data()
        exp = self._read_int_data()
        return decimal.Decimal(f"{mant}e{exp}")

    def _read_datetime(self) -> datetime.datetime:
        d = self._read_int_data()
        offset = 0
        has_tz_offset = d & 1
        has_not_msec = d & 2
        d >>= 2
        if has_tz_offset:
            offset = d & 0x7F
            if offset >= 128:
                offset -= 128  # sign extension
            d >>= 7
        f: float = d if has_not_msec else d / 1000
        f += ChainPack.SHV_EPOCH_SEC
        tzone = datetime.timezone(datetime.timedelta(minutes=offset * 15))
        return datetime.datetime.fromtimestamp(f, tzone)

    def _read_blob(self) -> bytes:
        dlen = self.read_uint_data()
        return self._read(dlen)

    def _read_string(self) -> str:
        slen = self.read_uint_data()
        if slen <= 0:
            return ""
        return self._read(slen).decode("utf-8")

    def _read_cstring(self) -> str:
        res = ""
        while True:
            b = self._read_byte()
            if b == ord("\\"):
                b = self._read_byte()
                if b == ord("\\"):
                    res += "\\"
                elif b == ord("0"):
                    res += "\0"
                else:
                    res += chr(b)
            else:
                if b == 0:
                    break  # end of string
                res += chr(b)
        return res

    def _read_list(self) -> SHVListType:
        lst = []
        while True:
            b = self._peek_byte()
            if b == ChainPack.CP_TERM:
                self._read_byte()
                break
            lst.append(self.read())
        return lst

    def _read_map(self) -> dict[str | int, SHVType]:
        mmap: dict[str | int, SHVType] = {}
        while True:
            b = self._peek_byte()
            if b == ChainPack.CP_TERM:
                self._read_byte()
                break
            key = self.read()
            if not isinstance(key, str | int):
                raise ValueError(f"Invalid Map key: {type(key)}")
            val = self.read()
            mmap[key] = val
        return mmap


class ChainPackWriter(commonpack.CommonWriter):
    """Write data in ChainPack format."""

    # Integer encoding:
    #  0 ...  7 bits  1  byte  |0|s|x|x|x|x|x|x|<-- LSB
    #  8 ... 14 bits  2  bytes |1|0|s|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 15 ... 21 bits  3  bytes |1|1|0|s|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 22 ... 28 bits  4  bytes |1|1|1|0|s|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 29+       bits  5+ bytes |1|1|1|1|n|n|n|n| |s|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| ... <-- LSB
    #                                         n ==  0 ->  4 bytes number (32 bit number)
    #                                         n ==  1 ->  5 bytes number
    #                                         n == 14 -> 18 bytes number
    #                                         n == 15 -> for future (number of bytes will be specified in next byte)

    @classmethod
    def _bytes_needed(cls, bit_len: int) -> int:
        """Calculate needed maximum number of bytes.

        Return max bit length >= bit_len, which can be encoded by same number of
        bytes number of bytes needed to encode bit_len.
        """
        if bit_len <= 28:
            cnt = ((bit_len - 1) // 7) + 1
        else:
            cnt = ((bit_len - 1) // 8) + 2
        return cnt or 1  # Always at least one byte is needed

    @classmethod
    def _expand_bit_len(cls, bit_len: int) -> int:
        byte_cnt = cls._bytes_needed(bit_len)
        if bit_len <= 28:
            ret = byte_cnt * (8 - 1) - 1
        else:
            ret = (byte_cnt - 1) * 8 - 1
        return ret

    def _write_uint_data_helper(self, num: int, bit_len: int) -> None:
        byte_cnt = self._bytes_needed(bit_len)
        data = bytearray(byte_cnt)
        for i in range(byte_cnt - 1, -1, -1):
            data[i] = num & 0xFF
            num >>= 8

        if bit_len <= 28:
            mask = 0xF0 << (4 - byte_cnt)
            data[0] &= ~mask
            mask = (mask << 1) & 0xFF
            data[0] |= mask
        else:
            data[0] = 0xF0 | (byte_cnt - 5)

        for i in range(0, byte_cnt):
            self._write(data[i])

    def write_meta(self, meta: SHVMetaType) -> None:  # noqa: D102
        self._write(ChainPack.CP_MetaMap)
        for k, v in meta.items():
            self.write(k)
            self.write(v)
        self._write(ChainPack.CP_TERM)

    def write_null(self) -> None:  # noqa: D102
        self._write(ChainPack.CP_Null)

    def write_bool(self, value: bool) -> None:  # noqa: D102
        self._write(ChainPack.CP_TRUE if value else ChainPack.CP_FALSE)

    def write_blob(self, value: bytes | bytearray) -> None:  # noqa: D102
        self._write(ChainPack.CP_Blob)
        self.write_uint_data(len(value))
        self._write(value)

    def write_string(self, value: str) -> None:  # noqa: D102
        bstring = value.encode("utf-8")
        self._write(ChainPack.CP_String)
        self.write_uint_data(len(bstring))
        self._write(bstring)

    def write_cstring(self, value: str) -> None:  # noqa: D102
        bstring = value.encode("utf-8")
        self._write(ChainPack.CP_CString)
        self._write(bstring)
        self._write(b"\0")

    def write_uint(self, value: int) -> None:  # noqa: D102
        if value < 64:
            self._write(value % 64)
        else:
            self._write(ChainPack.CP_UInt)
            self.write_uint_data(value)

    def write_uint_data(self, value: int) -> None:  # noqa: D102
        bitcnt = value.bit_length()
        self._write_uint_data_helper(value, bitcnt)

    def write_int(self, value: int) -> None:  # noqa: D102
        if 0 <= value < 64:
            self._write((value % 64) + 64)
        else:
            self._write(ChainPack.CP_Int)
            self.write_int_data(value)

    def write_int_data(self, value: int) -> None:  # noqa: D102
        num: int = abs(value)
        neg: bool = value < 0

        bitlen = num.bit_length()
        bitlen += 1  # add sign bit
        if neg:
            sign_pos = self._expand_bit_len(bitlen)
            sign_bit_mask = 1 << sign_pos
            num |= sign_bit_mask
        self._write_uint_data_helper(num, bitlen)

    def write_double(self, value: float) -> None:  # noqa: D102
        self._write(ChainPack.CP_Double)
        self._write(struct.pack("<d", value))  # little endian

    def write_decimal(self, value: decimal.Decimal) -> None:  # noqa: D102
        mantissa, exponent = decimal_rexp(value)
        self._write(ChainPack.CP_Decimal)
        self.write_int_data(mantissa)
        self.write_int_data(exponent)

    def write_list(self, value: SHVListType) -> None:  # noqa: D102
        self._write(ChainPack.CP_List)
        for val in value:
            self.write(val)
        self._write(ChainPack.CP_TERM)

    def write_map(self, value: SHVMapType) -> None:  # noqa: D102
        self._write(ChainPack.CP_Map)
        for k, v in value.items():
            self.write(k)
            self.write(v)
        self._write(ChainPack.CP_TERM)

    def write_imap(self, value: SHVIMapType) -> None:  # noqa: D102
        self._write(ChainPack.CP_IMap)
        for k, v in value.items():
            self.write(k)
            self.write(v)
        self._write(ChainPack.CP_TERM)

    def write_datetime(self, value: datetime.datetime) -> None:  # noqa: D102
        self._write(ChainPack.CP_DateTime)
        res = int(value.timestamp() * 1000) - (ChainPack.SHV_EPOCH_SEC * 1000)
        tzdelta = value.utcoffset()
        tzoff = int(tzdelta.total_seconds() // 60 // 15) if tzdelta is not None else 0
        if not -63 <= tzoff <= 63:
            raise ValueError(f"Invalid UTC offset value: {tzoff}")
        ms = res % 1000 == 0
        if ms:
            res //= 1000
        if tzoff != 0:
            res <<= 7
            res |= tzoff & 0x7F
        res <<= 2
        if tzoff != 0:
            res |= 1
        if ms:
            res |= 2
        self.write_int_data(res)



"""Implementation of SHV RPC message trasport between peers."""

from .abc import RpcClient, RpcServer
from .pipe import RpcClientPipe
from .stream import (
    RpcProtocolSerial,
    RpcProtocolSerialCRC,
    RpcProtocolStream,
    RpcTransportProtocol,
)
from .tcp import RpcClientTCP, RpcServerTCP
from .tty import RpcClientTTY, RpcServerTTY
from .unix import RpcClientUnix, RpcServerUnix
from .url import connect_rpc_client, create_rpc_server, init_rpc_client
from .ws import RpcClientWebSockets, RpcServerWebSockets, RpcServerWebSocketsUnix

__all__ = [
    "RpcClient",
    "RpcClientPipe",
    "RpcClientTCP",
    "RpcClientTTY",
    "RpcClientUnix",
    "RpcClientWebSockets",
    "RpcProtocolSerial",
    "RpcProtocolSerialCRC",
    "RpcProtocolStream",
    "RpcServer",
    "RpcServerTCP",
    "RpcServerTTY",
    "RpcServerUnix",
    "RpcServerWebSockets",
    "RpcServerWebSocketsUnix",
    "RpcTransportProtocol",
    "connect_rpc_client",
    "create_rpc_server",
    "init_rpc_client",
]



"""Connection over WebSockets."""

from __future__ import annotations

import abc
import asyncio
import collections.abc
import logging
import typing
import weakref

import websockets.client
import websockets.server

from .abc import RpcClient, RpcServer

logger = logging.getLogger(__name__)


class RpcClientWebSockets(RpcClient):
    """RPC connection to some SHV peer over WebSockets.

    :param uri: URI for the websocket connection.
    :param location: This is either hostname and path for the host connection or
      path to the Unix socket.
    :param port: Port used when connecting to the host. Keep it to ``-1`` if you
      are connecting to the Unix socket.
    """

    def __init__(self, location: str, port: int = -1) -> None:
        super().__init__()
        self.location = location
        self.port = port
        self._wsp: websockets.client.WebSocketClientProtocol | None = None
        self._close_task: asyncio.Task | None = None

    def __str__(self) -> str:
        if self.port == -1:
            return f"ws:{self.location}"
        return f"ws:{self.location}:{self.port}"

    async def _send(self, msg: bytes) -> None:
        if self._wsp is None:
            raise EOFError("Not connected")
        try:
            await self._wsp.send(msg)
        except (
            websockets.exceptions.ConnectionClosed,
            websockets.exceptions.ConnectionClosedOK,
            websockets.exceptions.ConnectionClosedError,
        ) as exc:
            raise EOFError from exc

    async def _receive(self) -> bytes:
        if self._wsp is None:
            raise EOFError("Not connected")
        try:
            # Ignore case when text frame is send to us for what ever reason.
            while True:
                if isinstance(res := await self._wsp.recv(), bytes):
                    return res
        except (
            websockets.exceptions.ConnectionClosed,
            websockets.exceptions.ConnectionClosedOK,
            websockets.exceptions.ConnectionClosedError,
        ) as exc:
            raise EOFError from exc

    async def reset(self) -> None:
        """Reset or establish the connection.

        This method not only resets the existing connection but primarilly
        estableshes the new one.
        """
        if not self.connected:
            if self.port != -1:
                self._wsp = await websockets.client.connect(
                    f"ws://{self.location}:{self.port}"
                )
            else:
                self._wsp = await websockets.client.unix_connect(self.location)
            self._close_task = None
            logger.debug("%s: Connected", self)
        else:
            await super().reset()

    @property
    def connected(self) -> bool:
        """Check if client is still connected."""
        return self._wsp is not None and self._wsp.open

    def _disconnect(self) -> None:
        if self._close_task is None and self._wsp is not None:
            self._close_task = asyncio.create_task(self._wsp.close())

    async def wait_disconnect(self) -> None:
        """Wait for the client's disconnection."""
        if self._wsp is not None:
            await self._wsp.wait_closed()
            if self._close_task is not None:
                await self._close_task


# TODO possibly implement RpcClientWebsocketsUnix if uri can't specify unix
# connection


class _RpcServerWebSockets(RpcServer):
    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], None | collections.abc.Awaitable[None]
        ],
    ) -> None:
        self.client_connected_cb = client_connected_cb
        """Callbact that is called when new client is connected."""
        self._server: websockets.server.WebSocketServer | None = None
        self._clients: weakref.WeakSet[_RpcServerWebSockets.Client] = weakref.WeakSet()

    @abc.abstractmethod
    async def _create_server(self) -> websockets.server.WebSocketServer:
        pass

    async def _client_connect(
        self, wsp: websockets.server.WebSocketServerProtocol
    ) -> None:
        client = self.Client(wsp, self)
        self._clients.add(client)
        res = self.client_connected_cb(client)
        if isinstance(res, collections.abc.Awaitable):
            await res
        # This coroutine must be blocked to not close the connection.
        await client.wait_disconnect()

    def is_serving(self) -> bool:
        return self._server is not None and self._server.is_serving()

    async def listen(self) -> None:
        if self._server is None:
            self._server = await self._create_server()
        await self._server.start_serving()

    async def listen_forewer(self) -> None:
        if self._server is None:
            self._server = await self._create_server()
        await self._server.serve_forever()

    def close(self) -> None:
        if self._server is not None:
            self._server.close()

    async def wait_closed(self) -> None:
        if self._server is not None:
            await self._server.wait_closed()

    def terminate(self) -> None:
        self.close()
        for client in self._clients:
            client.disconnect()

    async def wait_terminated(self) -> None:
        await self.wait_closed()
        res = await asyncio.gather(
            *(c.wait_disconnect() for c in self._clients),
            return_exceptions=True,
        )
        excs = [v for v in res if isinstance(v, BaseException)]
        if excs:
            if len(excs) == 1:
                raise excs[0]
            raise BaseExceptionGroup("", excs)

    class Client(RpcClient):
        def __init__(
            self,
            wsp: websockets.server.WebSocketServerProtocol,
            server: _RpcServerWebSockets,
        ) -> None:
            super().__init__()
            self.wsp = wsp
            self._close_task: asyncio.Task | None = None
            self._server = server

        async def _send(self, msg: bytes) -> None:
            try:
                await self.wsp.send(msg)
            except (
                websockets.exceptions.ConnectionClosed,
                websockets.exceptions.ConnectionClosedOK,
                websockets.exceptions.ConnectionClosedError,
            ) as exc:
                raise EOFError from exc

        async def _receive(self) -> bytes:
            try:
                # Ignore case when text frame is send to us for what ever reason.
                while True:
                    if isinstance(res := await self.wsp.recv(), bytes):
                        return res
            except (
                websockets.exceptions.ConnectionClosed,
                websockets.exceptions.ConnectionClosedOK,
                websockets.exceptions.ConnectionClosedError,
            ) as exc:
                raise EOFError from exc

        @property
        def connected(self) -> bool:
            # We are using not closed because open is false in the initial
            # connection phase.
            return not self.wsp.closed

        def _disconnect(self) -> None:
            if self._close_task is None and self.wsp is not None:
                self._close_task = asyncio.create_task(self.wsp.close())

        async def wait_disconnect(self) -> None:
            if self.wsp is not None:
                await self.wsp.wait_closed()
                if self._close_task is not None:
                    await self._close_task


class RpcServerWebSockets(_RpcServerWebSockets):
    """RPC server listening for WebSockets connections.

    :param client_connected_cb: Callback called when new connection is received.
    :param location: Hostname or IP address this server should listen on. The
      default ``None`` listens on all addresses.
    :param port: The port servers listens on.
    """

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], None | collections.abc.Awaitable[None]
        ],
        location: str | None = None,
        port: int = 8001,
    ) -> None:
        super().__init__(client_connected_cb)
        self.location = location
        """Host websockets server should listen on."""
        self.port = port
        """Port websocket server should listen on."""

    async def _create_server(self) -> websockets.server.WebSocketServer:
        return await websockets.server.serve(
            self._client_connect, self.location, self.port, start_serving=False
        )

    def __str__(self) -> str:
        location = (
            "locahost"
            if self.location is None
            else f"[{self.location}]"
            if ":" in self.location
            else self.location
        )
        return f"server.ws:{location}:{self.port}"

    class Client(_RpcServerWebSockets.Client):
        """RPC client for WebSockets server connection."""

        def __str__(self) -> str:
            address = self.wsp.remote_address
            location = f"[{address[0]}]" if ":" in address[0] else address[0]
            return f"ws:{location}:{address[1]}"


class RpcServerWebSocketsUnix(_RpcServerWebSockets):
    """RPC server listening for WebSockets connections on Unix socket.

    :param client_connected_cb: Callback called when new connection is received.
    :param location: Path where Unix socket will be created and server will
      listen on.
    """

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], None | collections.abc.Awaitable[None]
        ],
        location: str,
    ) -> None:
        super().__init__(client_connected_cb)
        self.location = location
        """Path to Unix socket server should listen on."""

    async def _create_server(self) -> websockets.server.WebSocketServer:
        return await websockets.server.unix_serve(self._client_connect, self.location)

    def __str__(self) -> str:
        return f"server.ws:{self.location}"

    class Client(_RpcServerWebSockets.Client):
        """RPC client for WebSockets server over Unix socket connection."""

        def __str__(self) -> str:
            return f"server.ws:{typing.cast(RpcServerWebSocketsUnix, self._server).location}"



"""Connection over TCP/IP."""

import asyncio
import collections.abc
import logging

from .abc import RpcClient
from .stream import (
    RpcClientStream,
    RpcProtocolStream,
    RpcServerStream,
    RpcTransportProtocol,
)

logger = logging.getLogger(__name__)


class RpcClientTCP(RpcClientStream):
    """RPC connection to some SHV peer over TCP/IP."""

    def __init__(
        self,
        location: str = "localhost",
        port: int = 3755,
        protocol: type[RpcTransportProtocol] = RpcProtocolStream,
    ) -> None:
        super().__init__(protocol)
        self.location = location
        self.port = port

    def __str__(self) -> str:
        location = f"[{self.location}]" if ":" in self.location else self.location
        return f"tcp:{location}:{self.port}"

    async def _open_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await asyncio.open_connection(self.location, self.port)


class RpcServerTCP(RpcServerStream):
    """RPC server listenting for SHV connections in TCP/IP."""

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], None | collections.abc.Awaitable[None]
        ],
        location: str | None = None,
        port: int = 3755,
        protocol: type[RpcTransportProtocol] = RpcProtocolStream,
    ) -> None:
        super().__init__(client_connected_cb, protocol)
        self.location = location
        self.port = port

    def __str__(self) -> str:
        location = (
            "locahost"
            if self.location is None
            else f"[{self.location}]"
            if ":" in self.location
            else self.location
        )
        return f"server.tcp:{location}:{self.port}"

    async def _create_server(self) -> asyncio.Server:
        # Note: The hack for '::' here is to allow really bind to all
        # interfaces as that is what this should do and  not just loopback.
        return await asyncio.start_server(
            self._client_connect,
            host=None if self.location == "::" else self.location,
            port=self.port,
        )

    async def listen(self) -> None:  # noqa: D102
        was_listening = self.is_serving()
        await super().listen()
        if not was_listening:
            logger.debug("%s: Listening for clients", self)

    def close(self) -> None:  # noqa: D102
        was_listening = self.is_serving()
        super().close()
        if was_listening and (self._server is None or not self._server.is_serving()):
            logger.debug("%s: No longer listening for clients", self)

    async def _client_connect(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        peername = writer.get_extra_info("peername")
        location = f"[{peername[0]}]" if ":" in peername[0] else peername[0]
        logger.debug("%s: New client %s:%d", self, location, peername[1])
        await super()._client_connect(reader, writer)

    class Client(RpcServerStream.Client):
        """RPC client for TCP server connection."""

        def __str__(self) -> str:
            peername = self._writer.get_extra_info("peername")
            location = f"[{peername[0]}]" if ":" in peername[0] else peername[0]
            return f"tcp:{location}:{peername[1]}"
